#!/usr/bin/env python
# cardinal_pythonlib/tools/pdf_to_booklet.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Command-line tool to make booklets from PDFs.**

RNC, 18 Nov 2017.

PURPOSE:

Take a PDF created with pdfnup or similar, with A4 sheets and two pages per
sheet, like this:

.. code-block:: none

    PDF page 1      +-----+-----+
                    |     |     |
                    |  1  |  2  |
                    |     |     |
                    +-----+-----+

    PDF page 2      +-----+-----+
                    |     |     |
                    |  3  |  4  |
                    |     |     |
                    +-----+-----+

    PDF page 3      +-----+-----+
                    |     |     |
                    |  5  |  6  |
                    |     |     |
                    +-----+-----+

and create a similar PDF but like this:

.. code-block:: none

    PDF page 1      +-----+-----+
                    |     |     |
                    |  6  |  1  |
                    |     |     |
                    +-----+-----+

    PDF page 2      +-----+-----+
                    |     |     |
                    |  1  |  2  |
                    |     |     |
                    +-----+-----+

    PDF page 3      +-----+-----+
                    |     |     |
                    |  1  |  2  |
                    |     |     |
                    +-----+-----+

so it can be printed double-sided and folded into an A5 booklet.

DEFINITIONS

- page = one side of a piece of paper BUT HERE, IN A BOOK CONTEXT, half that,
  i.e. what ends up as a book "page"
- pair = two pages, making up one side of a sheet/leaf
- sheet = one piece of paper (= leaf) (= 4 pages, here)

PRINTING

It's our job here to make pairs from pages, and to create a PDF where each
PDF page is a pair.

It's the printer's job to make sheets from pages. When printing in duplex,
you will need to use SHORT-EDGE BINDING (if you use long-edge binding, the
reverse sides will be inverted).

FURTHER THOUGHT 19 Nov 2017

We can, of course, support LONG-EDGE binding as well; that just requires
an extra step of rotating all the even-numbered pages from the preceding
step. Supported, as below.

"""

import argparse
import logging
import math
import os
import re
import shutil
import subprocess
import sys
import tempfile
from typing import List, NoReturn, Tuple
import unittest

from cardinal_pythonlib.cmdline import cmdline_quote
from cardinal_pythonlib.logs import BraceStyleAdapter, main_only_quicksetup_rootlogger  # noqa

log = BraceStyleAdapter(logging.getLogger(__name__))

CONVERT = "convert"
MUTOOL = "mutool"
PDFJAM = "pdfjam"
# PDFNUP = "pdfnup"  # interface to pdfjam, but too restrictive
PDFTK = "pdftk"

HELP_MISSING_IMAGEMAGICK = "Try 'sudo apt install imagemagick'"
HELP_MISSING_MUTOOL = "Try 'sudo apt install mupdf-tools'"
HELP_MISSING_PDFJAM = "Try 'sudo apt install pdfjam'"
HELP_MISSING_PDFTK = "Try 'sudo apt install pdftk'"

LATEX_PAPER_SIZE_A4 = "a4paper"

EXIT_SUCCESS = 0
EXIT_FAILURE = 1


# =============================================================================
# Calculate page sequence
# =============================================================================

def calc_n_sheets(n_pages: int) -> int:
    """
    How many sheets does this number of pages need, on the basis of 2 pages
    per sheet?
    """
    # NB PyCharm's type checker seems to think math.ceil() returns a float,
    # but it returns an int.
    # noinspection PyTypeChecker
    return math.ceil(n_pages / 2)


def calc_n_virtual_pages(n_sheets: int) -> int:
    """
    Converts #sheets to #pages, but rounding up to a multiple of 4.
    """
    if n_sheets % 2 == 0:
        return n_sheets * 2
    else:
        return (n_sheets + 1) * 2


def page_sequence(n_sheets: int, one_based: bool = True) -> List[int]:
    """
    Generates the final page sequence from the starting number of sheets.
    """
    n_pages = calc_n_virtual_pages(n_sheets)
    assert n_pages % 4 == 0
    half_n_pages = n_pages // 2
    firsthalf = list(range(half_n_pages))
    secondhalf = list(reversed(range(half_n_pages, n_pages)))
    # Seen from the top of an UNFOLDED booklet (e.g. a stack of paper that's
    # come out of your printer), "firsthalf" are on the right (from top to
    # bottom: recto facing up, then verso facing down, then recto, then verso)
    # and "secondhalf" are on the left (from top to bottom: verso facing up,
    # then recto facing down, etc.).
    sequence = []  # type: List[int]
    top = True
    for left, right in zip(secondhalf, firsthalf):
        if not top:
            left, right = right, left
        sequence += [left, right]
        top = not top
    if one_based:
        sequence = [x + 1 for x in sequence]
    log.debug("{} sheets => page sequence {!r}", n_sheets, sequence)
    return sequence


# =============================================================================
# PDF processor
# =============================================================================

def require(executable: str, explanation: str = "") -> None:
    """
    Ensures that the external tool is available.
    Asserts upon failure.
    """
    assert shutil.which(executable), "Need {!r} on the PATH.{}".format(
        executable, "\n" + explanation if explanation else "")


def run(args: List[str],
        get_output: bool = False,
        encoding: str = sys.getdefaultencoding()) -> Tuple[str, str]:
    """
    Run an external command +/- return the results.
    Returns a ``(stdout, stderr)`` tuple (both are blank strings if the output
    wasn't wanted).
    """
    printable = cmdline_quote(args)
    log.debug(f"Running external command: {printable}")
    if get_output:
        p = subprocess.run(args, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE, check=True)
        stdout, stderr = p.stdout.decode(encoding), p.stderr.decode(encoding)
    else:
        subprocess.check_call(args)
        stdout, stderr = "", ""
    return stdout, stderr


def get_page_count(filename: str) -> int:
    """
    How many pages are in a PDF?
    """
    log.debug("Getting page count for {!r}", filename)
    require(PDFTK, HELP_MISSING_PDFTK)
    stdout, _ = run([PDFTK, filename, "dump_data"], get_output=True)
    regex = re.compile(r"^NumberOfPages: (\d+)$", re.MULTILINE)
    m = regex.search(stdout)
    if m:
        return int(m.group(1))
    raise ValueError(f"Can't get PDF page count for: {filename!r}")


def make_blank_pdf(filename: str, paper: str = "A4") -> None:
    """
    NOT USED.
    Makes a blank single-page PDF, using ImageMagick's ``convert``.
    """
    # https://unix.stackexchange.com/questions/277892/how-do-i-create-a-blank-pdf-from-the-command-line  # noqa
    require(CONVERT, HELP_MISSING_IMAGEMAGICK)
    run([CONVERT, "xc:none", "-page", paper, filename])


def slice_pdf(input_filename: str, output_filename: str,
              slice_horiz: int, slice_vert: int) -> str:
    """
    Slice each page of the original, to convert to "one real page per PDF
    page". Return the output filename.
    """
    if slice_horiz == 1 and slice_vert == 1:
        log.debug("No slicing required")
        return input_filename  # nothing to do
    log.info("Slicing each source page mv into {} horizontally x {} vertically",
             slice_horiz, slice_vert)
    log.debug("... from {!r} to {!r}", input_filename, output_filename)
    require(MUTOOL, HELP_MISSING_MUTOOL)
    run([
        MUTOOL,
        "poster",
        "-x", str(slice_horiz),
        "-y", str(slice_vert),
        input_filename,
        output_filename
    ])
    return output_filename


def booklet_nup_pdf(input_filename: str, output_filename: str,
                    latex_paper_size: str = LATEX_PAPER_SIZE_A4) -> str:
    """
    Takes a PDF (e.g. A4) and makes a 2x1 booklet (e.g. 2xA5 per A4).
    The booklet can be folded like a book and the final pages will be in order.
    Returns the output filename.
    """
    log.info("Creating booklet")
    log.debug("... {!r} -> {!r}", input_filename, output_filename)
    require(PDFJAM, HELP_MISSING_PDFJAM)
    n_pages = get_page_count(input_filename)
    n_sheets = calc_n_sheets(n_pages)
    log.debug("{} pages => {} sheets", n_pages, n_sheets)
    pagenums = page_sequence(n_sheets, one_based=True)
    pagespeclist = [str(p) if p <= n_pages else "{}"
                    for p in pagenums]
    # ... switches empty pages to "{}", which is pdfjam notation for
    # an empty page.
    pagespec = ",".join(pagespeclist)
    pdfjam_tidy = True  # clean up after yourself?
    args = [
        PDFJAM,
        "--paper", latex_paper_size,
        "--landscape",
        "--nup", "2x1",
        "--keepinfo",  # e.g. author information
        "--outfile", output_filename,
        "--tidy" if pdfjam_tidy else "--no-tidy",
        "--",  # "no more options"
        input_filename, pagespec
    ]
    run(args)
    return output_filename


def rotate_even_pages_180(input_filename: str, output_filename: str) -> str:
    """
    Rotates even-numbered pages 180 degrees.
    Returns the output filename.
    """
    log.info("Rotating even-numbered pages 180 degrees for long-edge "
             "duplex printing")
    log.debug("... {!r} -> {!r}", input_filename, output_filename)
    require(PDFTK, HELP_MISSING_PDFTK)
    args = [
        PDFTK,
        "A=" + input_filename,  # give it handle 'A'
        # handles are one or more UPPER CASE letters
        "shuffle",
        "Aoddnorth",  # for 'A', keep odd pages as they are
        "Aevensouth",  # for 'A', rotate even pages 180 degrees
        "output", output_filename,
    ]
    run(args)
    return output_filename


def convert_to_foldable(input_filename: str,
                        output_filename: str,
                        slice_horiz: int,
                        slice_vert: int,
                        overwrite: bool = False,
                        longedge: bool = False,
                        latex_paper_size: str = LATEX_PAPER_SIZE_A4) -> bool:
    """
    Runs a chain of tasks to convert a PDF to a useful booklet PDF.
    """
    if not os.path.isfile(input_filename):
        log.warning("Input file does not exist or is not a file")
        return False
    if not overwrite and os.path.isfile(output_filename):
        log.error("Output file exists; not authorized to overwrite (use "
                  "--overwrite if you are sure)")
        return False
    log.info("Processing {!r}", input_filename)
    with tempfile.TemporaryDirectory() as tmpdir:
        log.debug("Using temporary directory {!r}", tmpdir)
        intermediate_num = 0

        def make_intermediate() -> str:
            nonlocal intermediate_num
            intermediate_num += 1
            return os.path.join(tmpdir,
                                f"intermediate_{intermediate_num}.pdf")

        # Run this as a chain, rewriting input_filename at each step:
        # Slice, if necessary.
        input_filename = slice_pdf(
            input_filename=input_filename,
            output_filename=make_intermediate(),
            slice_horiz=slice_horiz,
            slice_vert=slice_vert
        )
        # Make the final n-up
        input_filename = booklet_nup_pdf(
            input_filename=input_filename,
            output_filename=make_intermediate(),
            latex_paper_size=latex_paper_size
        )
        # Rotate?
        if longedge:
            input_filename = rotate_even_pages_180(
                input_filename=input_filename,
                output_filename=make_intermediate(),
            )
        # Done.
        log.info("Writing to {!r}", output_filename)
        shutil.move(input_filename, output_filename)
    return True


# =============================================================================
# Unit testing
# =============================================================================

class TestPdfToBooklet(unittest.TestCase):
    """
    Unit tests.
    """
    def test_sequence(self) -> None:
        for n_sheets in range(1, 8 + 1):
            log.info("{!r}", page_sequence(n_sheets=n_sheets, one_based=True))


# =============================================================================
# main
# =============================================================================

def main() -> NoReturn:
    """
    Command-line processor. See ``--help`` for details.
    """
    main_only_quicksetup_rootlogger(level=logging.DEBUG)
    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "input_file",
        help="Input PDF (which is not modified by this program)")
    parser.add_argument(
        "output_file",
        help="Output PDF")
    parser.add_argument(
        "--slice_horiz", type=int, default=1,
        help="Slice the input PDF first into this many parts horizontally")
    parser.add_argument(
        "--slice_vert", type=int, default=1,
        help="Slice the input PDF first into this many parts vertically")
    parser.add_argument(
        "--longedge", action="store_true",
        help="Create PDF for long-edge duplex printing, not short edge")
    parser.add_argument(
        "--overwrite", action="store_true",
        help="Allow overwriting of an existing output file")
    parser.add_argument(
        "--unittest", action="store_true",
        help="Run unit tests and exit (you must pass dummy values for "
             "input/output files to use these tests)")
    # ... because requiring dummy input/output filenames for unit testing
    # is less confusing for the majority of users than showing syntax in
    # which they are optional!
    args = parser.parse_args()

    if args.unittest:
        log.warning("Performing unit tests")
        # unittest.main() doesn't play nicely with argparse; they both
        # use sys.argv by default (and we end up with what looks like garbage
        # from the argparse help facility); but this works:
        unittest.main(argv=[sys.argv[0]])
        sys.exit(EXIT_SUCCESS)

    success = convert_to_foldable(
        input_filename=os.path.abspath(args.input_file),
        output_filename=os.path.abspath(args.output_file),
        slice_horiz=args.slice_horiz,
        slice_vert=args.slice_vert,
        overwrite=args.overwrite,
        longedge=args.longedge
    )
    sys.exit(EXIT_SUCCESS if success else EXIT_FAILURE)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# cardinal_pythonlib/tcpipconst.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Constants for use with TCP/IP.**

"""


# =============================================================================
# Common TCP/IP ports
# =============================================================================

class Ports(object):
    """
    Common TCP/UDP ports.

    - https://en.wikipedia.org/wiki/Port_(computer_networking)
    - https://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers
    """

    # -------------------------------------------------------------------------
    # System (well-known) ports are in the range 0-1023.
    # -------------------------------------------------------------------------

    FTP_DATA = 20  # not for initiating connections
    FTP_CONTROL = 21  # for initiating connections
    FTP = FTP_CONTROL
    SSH = 22
    SFTP = SSH  # SFTP is FTP over SSH
    TELNET = 23
    HTTP = 80
    NTP = 123
    HTTPS = 443

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # E-mail is a special mess:
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # As of RFC8314 (2018), the recommendations are:
    #
    # - Connect via "implicit TLS" ports (always encrypted), rather than
    #   connecting to the "cleartext" port and negotiating TLS using the
    #   "STARTTLS" command.
    #
    #   - I think that means: use 465 for SMTP, 993 for IMAP, 995 for POP3.
    #
    # However, 587 was always the official e-mail submission port. And RFC8134
    # is a standards-track document, not an accepted standard
    # (https://www.rfc-editor.org/info/rfc8314). Currently RFC8314 says that
    # servers SHOULD support both 587 and 465.
    #
    # Older sources will say that 587 is preferred for SMTP. See
    # https://serverfault.com/questions/1064955/why-is-port-587-preferred-over-port-465-in-smtp  # noqa

    SMTP = 25  # its formal IANA name;
    # ... https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xhtml?search=25  # noqa
    SMTP_UNENCRYPTED = 25  # https://datatracker.ietf.org/doc/html/rfc5321
    SMTP_MTA = SMTP
    # ... MTA = message transfer agent = for server-to-server transmission
    # ... receives from an MUA (mail user agent = e-mail client)

    POP3 = 110  # formal name
    POP3_UNENCRYPTED = POP3

    IMAP = 143  # formal name
    IMAP_UNENCRYPTED = IMAP

    SUBMISSIONS = 465  # formal name; message submission over TLS
    # ... "submission-s[ecure]", not plural of submission.
    SMTPS = SUBMISSIONS
    # ... https://datatracker.ietf.org/doc/html/rfc8314
    # ... https://stackoverflow.com/questions/15796530/what-is-the-difference-between-ports-465-and-587  # noqa
    # ... requires negotiation of TLS/SSL at connection setup
    # Was initially intended for another purpose (URL Rendezvous Directory for
    # SSM), but was co-opted for e-mail and formalized by RFC8314.
    SMTP_PREFERRED = SUBMISSIONS
    # ... this is the best for client-to-server transmission (per RFC8314)

    SUBMISSION = 587  # formal name; message submission
    SMTP_MSA = SUBMISSION
    # ... MSA = message submission agent = client-to-server
    # ... can
    # ... https://datatracker.ietf.org/doc/html/rfc8314
    # ... https://stackoverflow.com/questions/15796530/what-is-the-difference-between-ports-465-and-587  # noqa
    # ... requires SMTP AUTH (authentication);
    #     https://en.wikipedia.org/wiki/SMTP_Authentication
    # ... uses STARTTLS if one chooses to negotiate TLS

    IMAPS = 993  # formal name; IMAP over TLS
    IMAP_ENCRYPTED = IMAPS

    POP3S = 995  # formal name; POP3 over TLS
    POP3_ENCRYPTED = POP3S

    # -------------------------------------------------------------------------
    # Registered ports are in the range 1024 to 49151.
    # -------------------------------------------------------------------------

    MS_SQL_S = 1433  # formal name; a database
    MICROSOFT_SQL_SERVER = MS_SQL_S  # a database
    HL7 = 2575  # formal name; HL7 messaging
    HL7_MLLP = HL7  # HL7 MLLP = Minimum Lower Layer Protocol
    MYSQL = 3306  # a database
    WHISKER = 3233  # a behavioural research control system
    POSTGRESQL = 5432  # a database
    AMQP = 5672  # Advanced Message Queuing Protocol

    # -------------------------------------------------------------------------
    # Dynamic ports are in the range 49152 to 65535.
    # -------------------------------------------------------------------------

    # -------------------------------------------------------------------------
    # Some non-standard ports:
    # -------------------------------------------------------------------------

    FLOWER_NONSTANDARD = 5555  # default internal port for Flower, https://flower.readthedocs.io/  # noqa
    ALTERNATIVE_HTTP_NONSTANDARD = 8000
    RABBITMQ_ADMIN_NONSTANDARD = 15672  # default internal port for RabbitMQ admin interface; https://www.rabbitmq.com/  # noqa


class UriSchemes:
    """
    Common URI schemes.
    See https://en.wikipedia.org/wiki/List_of_URI_schemes.
    """
    FTP = "ftp"
    HTTP = "http"
    HTTPS = "https"
    IMAP = "imap"
    MAILTO = "mailto"
    SFTP = "sftp"
    SMS = "sms"
    SSH = "ssh"
    TEL = "tel"  # telephone
    TELNET = "telnet"



#!/usr/bin/env python
# cardinal_pythonlib/chebi.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to assist with the ChEBI database.**

ChEBI: Chemical Entities of Biological Interest (ChEBI) database from
EMBL-EBI (European Molecular Biology Laboratory / European Bioinformatics
Institute).

See https://www.ebi.ac.uk/chebi/

Examples:

.. code-block:: python

    cardinalpythonlib_chebi test

    cardinalpythonlib_chebi search citalopram
    cardinalpythonlib_chebi search citalopram --exact_search
    cardinalpythonlib_chebi search zopiclone
    cardinalpythonlib_chebi search zopiclone --exact_search
    cardinalpythonlib_chebi search zopiclone --exact_match
    cardinalpythonlib_chebi search salicylic --inexact_search

    cardinalpythonlib_chebi describe citalopram simvastatin --exact_match

    cardinalpythonlib_chebi ancestors citalopram simvastatin

Then try this syntax:

.. code-block:: bash

    cardinalpythonlib_chebi categorize \
        --entities entities.txt \
        --entity_synonyms entity_synonyms.txt \
        --categories categories.txt \
        --category_synonyms category_synonyms.txt \
        --manual_categories manual_categories.txt \
        --results results.csv

using files like these:

.. code-block:: none

    # entities.txt
    # Things to classify.

    agomelatine
    aspirin
    citalopram
    simvastatin

.. code-block:: none

    # entity_synonyms.txt
    # Renaming of entities prior to lookup.
    # Find these via "cardinalpythonlib_chebi search ..." or Google with "CHEBI".
    
    aspirin, acetylsalicylic acid

.. code-block:: none

    # categories.txt
    # Categories to detect, in order of priority (high to low).
    
    serotonin reuptake inhibitor
    antidepressant

    antilipemic drug

    non-steroidal anti-inflammatory drug

.. code-block:: none

    # category_synonyms.txt
    # Categories that are equivalent but ChEBI doesn't know. 
    
    glucagon-like peptide-1 receptor agonist, hypoglycemic agent

.. code-block:: none

    # manual_categories.txt
    # Categorizations that ChEBI doesn't know. 
    
    agomelatine, antidepressant

"""  # noqa

import argparse
import csv
import logging
from typing import List, Generator, Optional, Sequence, Set, Tuple, Union

from appdirs import user_cache_dir
try:
    # noinspection PyPackageRequirements
    from libchebipy import (
        ChebiEntity,
        Relation,
        search,
        set_download_cache_path,
    )
except ImportError:
    raise ImportError("Cannot import libchebipy; try the command: pip install libChEBIpy")  # noqa

from cardinal_pythonlib.file_io import (
    gen_lines_without_comments,
    get_lines_without_comments,
)
from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger
from cardinal_pythonlib.version_string import VERSION_STRING

log = logging.getLogger(__name__)

# =============================================================================
# Constants
# =============================================================================

DEFAULT_CACHE_PATH = user_cache_dir("chebi")
DEFAULT_ANCESTOR_RELATIONSHIPS = ["has_role", "is_a"]  # both are helpful

DEFAULT_EXACT_SEARCH = False
DEFAULT_EXACT_MATCH = False


# =============================================================================
# Hashable version of ChebiEntity
# =============================================================================

_CHEBI_ID_PREFIX = "CHEBI:"


def get_chebi_id_number_str(entity: ChebiEntity) -> str:
    """
    Returns the CHEBI ID number as a string.

    Args:
        entity:
            a :class:`libchebipy.ChebiEntity`
    """
    return entity.get_id().replace(_CHEBI_ID_PREFIX, "")


def get_chebi_id_number(entity: ChebiEntity) -> int:
    """
    Returns the CHEBI ID number as an integer.

    Args:
        entity:
            a :class:`libchebipy.ChebiEntity`
    """
    return int(get_chebi_id_number_str(entity))


class HashableChebiEntity(ChebiEntity):
    """
    Hashable version of :class:`libchebipy.ChebiEntity`.
    """

    @classmethod
    def from_chebi_entity(cls, entity: ChebiEntity) -> "HashableChebiEntity":
        id_number_str = get_chebi_id_number_str(entity)
        return HashableChebiEntity(id_number_str)

    def get_id_number_str(self) -> str:
        return get_chebi_id_number_str(self)

    def get_id_number(self) -> int:
        return get_chebi_id_number(self)

    def __eq__(self, other: Union[str, int, "HashableChebiEntity"]) -> bool:
        if isinstance(other, str):
            return other == self.get_name()
        elif isinstance(other, int):
            return other == self.get_id_number()
        else:
            # noinspection PyUnresolvedReferences
            return self.get_id_number() == other.get_id_number()

    def __hash__(self) -> int:
        return self.get_id_number()


# =============================================================================
# Descriptions of a ChebiEntity
# =============================================================================

def brief_description(entity: ChebiEntity) -> str:
    """
    Args:
        entity:
            a :class:`ChebiEntity`

    Returns:
        str: name and ID

    """
    return f"{entity.get_name()} ({entity.get_id()})"


# =============================================================================
# Searching ChEBI
# =============================================================================

def get_entity(chebi_id: Union[int, str]) -> ChebiEntity:
    """
    Fetch a ChEBI entity by its ID.

    Args:
        chebi_id:
            integer ChEBI ID like ``15903``, or string ID like ``'15903'``,
            or string ID like ``'CHEBI:15903'``.
    """
    chebi_id = str(chebi_id)  # ignore buggy demo code; int not OK
    log.debug(f"Looking up ChEBI ID: {chebi_id}")
    return ChebiEntity(chebi_id)


def search_entities(search_term: Union[int, str],
                    exact_search: bool = DEFAULT_EXACT_SEARCH,
                    exact_match: bool = DEFAULT_EXACT_MATCH) \
        -> List[ChebiEntity]:
    """
    Search for ChEBI entities.

    Case-insensitive.

    Args:
        search_term:
            String or integer to search for.
        exact_search:
            The ``exact`` parameter to :func:`libchebipy.search`.
        exact_match:
            Ensure that the name of the result exactly matches the search term.
            Example: an exact search for "zopiclone" gives both "zopiclone
            (CHEBI:32315)" and "(5R)-zopiclone (CHEBI:53762)"; this option
            filters to the first.
    """
    log.debug(f"Searching for {search_term!r} "
              f"(exact_search={exact_search}, exact_match={exact_match})")
    results = search(search_term, exact=exact_search)
    log.debug(f"libchebipy.search({search_term!r}, exact={exact_search}) "
              f"-> {results!r}")
    if exact_match:
        if isinstance(search_term, int):
            results = [r for r in results
                       if get_chebi_id_number(r) == search_term]
        else:
            assert isinstance(search_term, str)
            results = [r for r in results
                       if r.get_name().lower() == search_term.lower()]
    log.debug(f"search_entities({search_term!r}, exact_search={exact_search}, "
              f"exact_match={exact_match}) -> {results!r}")
    return results


# =============================================================================
# Describing ChEBI entries
# =============================================================================

def describe_entity(entity: ChebiEntity) -> None:
    """
    Test function to describe a ChEBI entity.

    Args:
        entity:
            a :class:`ChebiEntity`
    """
    name = entity.get_name()

    out_lines = []  # type: List[str]
    for other in entity.get_outgoings():
        target = ChebiEntity(other.get_target_chebi_id())
        out_lines.append(
            f"    • {name} {other.get_type()} {brief_description(target)}")

    in_lines = []  # type: List[str]
    for other in entity.get_incomings():
        target = ChebiEntity(other.get_target_chebi_id())
        in_lines.append(
            f"    • {brief_description(target)} {other.get_type()} {name}")

    lines = (
        [entity.get_name(), f"  ► OUTGOING ({len(out_lines)})"] +
        out_lines +
        [f"  ► INCOMING ({len(in_lines)})"] +
        in_lines
    )
    report = "\n".join(lines)
    log.info(f"{entity.get_id()}:\n{report}")


def search_and_describe(search_term: Union[int, str],
                        exact_search: bool = DEFAULT_EXACT_SEARCH,
                        exact_match: bool = DEFAULT_EXACT_MATCH) -> None:
    """
    Search for a ChEBI term and describe it to the log.

    Args:
        search_term: search term
        exact_search: exact search?
        exact_match: exact match?
    """
    entities = search_entities(search_term, exact_search=exact_search,
                               exact_match=exact_match)
    for entity in entities:
        describe_entity(entity)


def search_and_describe_multiple(
        search_terms: List[Union[int, str]],
        exact_search: bool = DEFAULT_EXACT_SEARCH,
        exact_match: bool = DEFAULT_EXACT_MATCH) -> None:
    """
    Search for ChEBI terms; describe matching entries to the log.

    Args:
        search_terms: search term(s)
        exact_search: exact search?
        exact_match: exact match?
    """
    for search_term in search_terms:
        search_and_describe(search_term, exact_search=exact_search,
                            exact_match=exact_match)


def search_and_list(search_term: Union[int, str],
                    exact_search: bool = DEFAULT_EXACT_SEARCH,
                    exact_match: bool = DEFAULT_EXACT_MATCH) -> None:
    """
    Search for a ChEBI term; print matching entries to the log.

    Args:
        search_term: search term
        exact_search: exact search?
        exact_match: exact match?
    """
    entities = search_entities(search_term, exact_search=exact_search,
                               exact_match=exact_match)
    lines = [f"– {brief_description(entity)}" for entity in entities]
    report = "\n".join(lines)
    log.info(f"Results:\n{report}")


def search_and_list_multiple(search_terms: List[Union[int, str]],
                             exact_search: bool = DEFAULT_EXACT_SEARCH,
                             exact_match: bool = DEFAULT_EXACT_MATCH) -> None:
    """
    Search for ChEBI terms; print matching entries to the log.

    Args:
        search_terms: search term(s)
        exact_search: exact search?
        exact_match: exact match?
    """
    for search_term in search_terms:
        search_and_list(search_term, exact_search=exact_search,
                        exact_match=exact_match)


# =============================================================================
# Ancestors and descendants of ChEBI entities
# =============================================================================

def gen_ancestor_info(entity: ChebiEntity,
                      relationships: List[str] = None,
                      max_generations: int = None,
                      starting_generation_: int = 0,
                      seen_: Set[HashableChebiEntity] = None) \
        -> Generator[Tuple[HashableChebiEntity, str, int], None, None]:
    """
    Retrieves all ancestors ("outgoing" links).

    Args:
        entity:
            starting entity
        relationships:
            list of valid relationship types, e.g. "has_role"
        max_generations:
            maximum number of generations to pursue, or ``None`` for unlimited
        starting_generation_:
            for internal use only, for recursion
        seen_:
            for internal use only, for recursion

    Returns:
        list: of tuples ``entity, relationship, n_generations_above_start``
    """
    if max_generations is not None and starting_generation_ >= max_generations:
        return
    assert starting_generation_ == 0 or seen_ is not None
    seen_ = seen_ or set()  # type: Set[HashableChebiEntity]
    relationships = relationships or DEFAULT_ANCESTOR_RELATIONSHIPS
    log.debug(f"Finding ancestors of {brief_description(entity)} "
              f"(generation {starting_generation_}) "
              f"via relationships {relationships!r}")
    for rel in entity.get_outgoings():  # type: Relation
        if rel.get_type() in relationships:
            target = HashableChebiEntity(rel.get_target_chebi_id())
            log.debug(f"... found {brief_description(target)}")
            if target in seen_:
                # log.debug(f"Skipping {target!r}")
                continue
            seen_.add(target)
            yield target, rel.get_type(), starting_generation_ + 1
            yield from gen_ancestor_info(
                entity=target,
                relationships=relationships,
                starting_generation_=starting_generation_ + 1,
                seen_=seen_,
            )


def gen_ancestors(entity: ChebiEntity,
                  relationships: List[str] = None,
                  max_generations: int = None) \
        -> Generator[HashableChebiEntity, None, None]:
    """
    Generates ancestors as per :func:`gen_ancestor_info`, without relationship
    or generation info.
    """
    for (ancestor,
         relationship,
         generation) in gen_ancestor_info(entity,
                                          relationships,
                                          max_generations):
        yield ancestor


def report_ancestors(entity: ChebiEntity,
                     relationships: List[str] = None,
                     max_generations: int = None) -> None:
    """
    Fetches and reports on ancestors of an entity, e.g. via "is_a"
    relationships. See :func:`gen_ancestor_info`.
    """
    relationships = relationships or DEFAULT_ANCESTOR_RELATIONSHIPS
    ancestors = list(gen_ancestor_info(
        entity=entity,
        relationships=relationships,
        max_generations=max_generations))
    lines = [f"{entity.get_name()} ({entity.get_id()})"]
    for ancestor, relationship, generation in ancestors:
        prefix = "  " * generation
        lines.append(f"{prefix}► {relationship} "
                     f"{brief_description(ancestor)} [{generation}]")
    report = "\n".join(lines)
    log.info(f"Ancestors via {relationships!r}:\n{report}")


def report_ancestors_multiple(entity_names: List[str],
                              relationships: List[str] = None,
                              max_generations: int = None) -> None:
    """
    Looks up entities, then reports on ancestors.
    Fetches and reports on ancestors of an entity, e.g. via "is_a"
    relationships. See :func:`gen_ancestor_info`.
    """
    log.debug(f"Using ancestor relationships: {relationships!r}")
    log.debug(f"Using max_generations: {max_generations!r}")
    for entity_name in entity_names:
        for entity in search_entities(entity_name):
            report_ancestors(entity, relationships, max_generations)


# =============================================================================
# Testing
# =============================================================================

def testfunc1() -> None:
    """
    Test ChEBI interface.
    """
    log.warning("Testing: describe beta-D-glucose")
    beta_d_glucose = get_entity(15903)
    describe_entity(beta_d_glucose)
    # Cross-check:
    # https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:15903
    # ... correct.

    log.warning("Testing: describe things like 'citalopram'")
    search_and_describe("citalopram", exact_search=False)

    log.warning("Testing: show ancestors of citalopram")
    citalopram = get_entity(3723)
    report_ancestors(citalopram)
    # https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:3723

    sri = "serotonin uptake inhibitor"
    log.warning(f"Testing: search/list {sri!r}")
    search_and_list(sri)
    log.warning(f"Testing: search/describe {sri!r}")
    search_and_describe(sri)


# =============================================================================
# Mapping terms via dictionaries
# =============================================================================

class CaseInsensitiveDict(dict):
    """
    Case-insensitive dictionary for strings; see
    https://stackoverflow.com/questions/2082152/case-insensitive-dictionary
    """
    def __setitem__(self, key: str, value: str) -> None:
        # https://docs.python.org/3/reference/datamodel.html#object.__setitem__
        super().__setitem__(key.lower(), value)

    def __contains__(self, key: str) -> bool:
        # https://docs.python.org/3/reference/datamodel.html#object.__contains__
        return super().__contains__(key.lower())

    def __getitem__(self, key: str) -> str:
        # https://docs.python.org/3/reference/datamodel.html#object.__getitem__
        return super().__getitem__(key.lower())


def read_dict(filename: str) -> CaseInsensitiveDict:
    """
    Reads a filename that may have comments but is otherwise in the format

    .. code-block:: none

        a1, b1
        a2, b2
        ...

    Args:
        filename:
            filename to read

    Returns:
        dict: mapping the first column (converted to lower case) to the second
        (case left intact).

    """
    d = CaseInsensitiveDict()
    for line in gen_lines_without_comments(filename):
        parts = [p.strip() for p in line.split(",")]
        if len(parts) == 2:
            a = parts[0]
            b = parts[1]
            d[a] = b
        else:
            log.error(f"Bad CSV-pair line: {line!r}")
    return d


def translate(term: str, mapping: CaseInsensitiveDict) -> Tuple[str, bool]:
    """
    Translates a term through a dictionary. If the term (once converted to
    lower case) is in the dictionary (see :func:`read_dict`), the mapped term
    is returned; otherwise the original search term is returned.

    Args:
        term:
            term to look up
        mapping:
            the mapping dictionary

    Returns:
        tuple: result (str), renamed? (bool)

    """
    result = mapping.get(term, term)
    return result, result != term


# =============================================================================
# Categorizing drugs
# =============================================================================

def get_category(entity_name: str,
                 categories: Sequence[str],
                 entity_synonyms: CaseInsensitiveDict = None,
                 category_synonyms: CaseInsensitiveDict = None,
                 manual_categories: CaseInsensitiveDict = None,
                 relationships: List[str] = None) -> Optional[str]:
    """

    Args:
        entity_name:
            name of entity to categorize
        categories:
            permissible categories (earlier preferable to later)
        entity_synonyms:
            map to rename entities
        category_synonyms:
            mapping of categories to other (preferred) categories
        manual_categories:
            manual overrides mapping entity to category
        relationships:
            list of valid relationship types defining ancestry, e.g. "has_role"

    Returns:
        chosen category, or ``None`` if none found
    """
    entity_synonyms = entity_synonyms or CaseInsensitiveDict()
    category_synonyms = category_synonyms or CaseInsensitiveDict()
    manual_categories = manual_categories or CaseInsensitiveDict()
    relationships = relationships or DEFAULT_ANCESTOR_RELATIONSHIPS

    log.debug(f"get_category: entity_name={entity_name!r}")

    # Manual override for original name?
    if entity_name in manual_categories:
        category, _ = translate(manual_categories[entity_name],
                                category_synonyms)
        log.debug(f"Manual categorization: {entity_name} → {category}")
        return category

    # Renamed?
    entity_name, renamed = translate(entity_name, entity_synonyms)

    # Manual override for renamed entity?
    if renamed:
        if entity_name in manual_categories:
            category, _ = translate(manual_categories[entity_name],
                                    category_synonyms)
            log.debug(f"Manual categorization: {entity_name} → {category}")
            return category

    # Find entity
    entities = search_entities(entity_name,
                               exact_search=True,
                               exact_match=True)
    if len(entities) == 0:
        log.warning(f"No entity found for {entity_name!r}")
        return None
    if len(entities) > 1:
        descriptions = "; ".join(brief_description(e) for e in entities)
        log.warning(f"Multiple entities found for {entity_name!r}; "
                    f"using the first. They were:\n{descriptions}")
    entity = entities[0]

    # Find category
    ancestors = list(gen_ancestors(entity, relationships=relationships))
    ancestor_categories = [
        translate(a.get_name(), category_synonyms)[0]
        for a in ancestors
    ]
    # log.debug(f"ancestor_categories: {ancestor_categories!r}")
    for category in categories:  # implements category order
        category, _ = translate(category, category_synonyms)
        if category in ancestor_categories:
            return category
    return None


def categorize_from_file(entity_filename: str,
                         category_filename: str,
                         results_filename: str,
                         entity_synonyms_filename: str = None,
                         category_synonyms_filename: str = None,
                         manual_categories_filename: str = None,
                         relationships: List[str] = None,
                         output_dialect: str = "excel",
                         headers: bool = True) -> None:
    """
    Categorizes entities.

    Args:
        entity_filename:
            input filename, one entity per line
        category_filename:
            filename containing permissible categories, one per line
            (earlier preferable to later)
        results_filename:
            output filename for CSV results
        entity_synonyms_filename
            Name of CSV file (with optional # comments) containing synonyms
            in the format ``entity_from, entity_to``.
        category_synonyms_filename:
            Name of CSV file (with optional # comments) containing synonyms
            in the format ``category_from, categoryto``.
        manual_categories_filename:
            Name of CSV file (with optional # comments) containing manual
            categorizations in the format ``entity, category``.
        relationships:
            list of valid relationship types defining ancestry, e.g. "has_role"
        output_dialect:
            CSV output dialect
        headers:
            add CSV headers?
    """
    relationships = relationships or DEFAULT_ANCESTOR_RELATIONSHIPS
    log.info(f"Using ancestor relationships {relationships!r}")

    log.info(f"Reading categories from {category_filename}")
    categories = get_lines_without_comments(category_filename)

    if entity_synonyms_filename:
        log.info(f"Reading entity synonyms from {entity_synonyms_filename}")
        entity_synonyms = read_dict(entity_synonyms_filename)
    else:
        entity_synonyms = CaseInsensitiveDict()
    log.debug(f"Using entity synonyms: {entity_synonyms!r}")

    if category_synonyms_filename:
        log.info(f"Reading category synonyms from {category_synonyms_filename}")
        category_synonyms = read_dict(category_synonyms_filename)
    else:
        category_synonyms = CaseInsensitiveDict()
    log.debug(f"Using category synonyms: {category_synonyms!r}")

    if manual_categories_filename:
        log.info(f"Reading manual categories from {manual_categories_filename}")  # noqa
        manual_categories = read_dict(manual_categories_filename)
    else:
        manual_categories = CaseInsensitiveDict()
    log.debug(f"Using manual categories: {manual_categories!r}")

    log.info(f"Writing to {results_filename!r}")
    entities_seen = set()  # type: Set[str]
    with open(results_filename, "w") as outfile:
        writer = csv.writer(outfile, dialect=output_dialect)
        if headers:
            writer.writerow(["entity", "category"])
        log.info(f"Reading entities from {entity_filename}")
        for entity_name in gen_lines_without_comments(entity_filename):
            entity_name_lower = entity_name.lower()
            if entity_name_lower in entities_seen:
                log.warning(f"Ignoring duplicate: {entity_name!r}")
                continue
            entities_seen.add(entity_name_lower)
            category = get_category(
                entity_name=entity_name,
                categories=categories,
                entity_synonyms=entity_synonyms,
                category_synonyms=category_synonyms,
                manual_categories=manual_categories,
                relationships=relationships
            ) or ""
            if category:
                log.debug(f"{entity_name} → {category}")
            else:
                log.error(f"No category found for {entity_name!r}")
            writer.writerow([entity_name, category])


# =============================================================================
# Main
# =============================================================================

def main() -> None:
    """
    Command-line entry point.
    """
    # Parser
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--cachepath", type=str, default=DEFAULT_CACHE_PATH,
        help="Cache path for ChEBI files"
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Be verbose"
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="Valid subcommands",
        dest="command"
    )
    subparsers.required = True

    def add_exact(p: argparse.ArgumentParser) -> None:
        p.add_argument(
            "--exact_search", dest="exact_search", action="store_true",
            help="Search using exact term")
        p.add_argument(
            "--inexact_search", dest="exact_search", action="store_false",
            help="Search allowing inexact matches")
        p.set_defaults(exact_search=DEFAULT_EXACT_SEARCH)
        p.add_argument(
            "--exact_match", dest="exact_match", action="store_true",
            help="Return results for exact term only")
        p.add_argument(
            "--inexact_match", dest="exact_match", action="store_false",
            help="Return results allowing inexact matches")
        p.set_defaults(exact_match=DEFAULT_EXACT_MATCH)

    def add_entities(p: argparse.ArgumentParser) -> None:
        p.add_argument(
            "entity", type=str, nargs="+",
            help="Entity or entities to search for"
        )

    # -------------------------------------------------------------------------
    # Test
    # -------------------------------------------------------------------------
    parser_test = subparsers.add_parser(
        "test",
        help="Run some simple tests"
    )
    parser_test.set_defaults(func=lambda args: testfunc1())

    # -------------------------------------------------------------------------
    # Search
    # -------------------------------------------------------------------------
    parser_search = subparsers.add_parser(
        "search",
        help="Search for an entity in the ChEBI database"
    )
    add_entities(parser_search)
    add_exact(parser_search)
    parser_search.set_defaults(func=lambda args: search_and_list_multiple(
        search_terms=args.entity,
        exact_search=args.exact_search,
        exact_match=args.exact_match,
    ))

    # -------------------------------------------------------------------------
    # Describe
    # -------------------------------------------------------------------------
    parser_describe = subparsers.add_parser(
        "describe",
        help="Describe an entity/entities in the ChEBI database"
    )
    add_entities(parser_describe)
    add_exact(parser_describe)
    parser_describe.set_defaults(func=lambda args: search_and_describe_multiple(  # noqa
        search_terms=args.entity,
        exact_search=args.exact_search,
        exact_match=args.exact_match,
    ))

    # -------------------------------------------------------------------------
    # Ancestors
    # -------------------------------------------------------------------------
    parser_ancestors = subparsers.add_parser(
        "ancestors",
        help="Show ancestors of an entity/entities in the ChEBI database"
    )
    add_entities(parser_ancestors)
    parser_ancestors.add_argument(
        "--relationships", type=str, nargs="+",
        default=DEFAULT_ANCESTOR_RELATIONSHIPS,
        help="Relationship types that define an ancestor"
    )
    parser_ancestors.add_argument(
        "--max_generations", type=int, default=None,
        help="Number of generations to search, or None for unlimited"
    )
    parser_ancestors.set_defaults(func=lambda args: report_ancestors_multiple(
        entity_names=args.entity,
        relationships=args.relationships,
        max_generations=args.max_generations,
    ))

    # -------------------------------------------------------------------------
    # Categorize
    # -------------------------------------------------------------------------
    parser_categorize = subparsers.add_parser(
        "categorize",
        help="Categorize a list of drugs."
    )
    parser_categorize.add_argument(
        "--entities", type=str, required=True,
        help="Input file, one entity (e.g. drug) name per line."
    )
    parser_categorize.add_argument(
        "--categories", type=str, required=True,
        help="Name of file containing categories, one per line "
             "(earlier categories preferred to later)."
    )
    parser_categorize.add_argument(
        "--entity_synonyms", type=str, default=None,
        help="Name of CSV file (with optional # comments) containing synonyms "
             "in the format 'entity_from, entity_to'"
    )
    parser_categorize.add_argument(
        "--category_synonyms", type=str, default=None,
        help="Name of CSV file (with optional # comments) containing synonyms "
             "in the format 'category_from, category_to'. The translation is "
             "applied to ChEBI categories before matching. For example you "
             "can map 'EC 3.1.1.7 (acetylcholinesterase) inhibitor' to "
             "'acetylcholinesterase inhibitor' and then use only "
             "'acetylcholinesterase inhibitor' in your category file."
    )
    parser_categorize.add_argument(
        "--manual_categories", type=str, default=None,
        help="Name of CSV file (with optional # comments) containing manual "
             "categorizations in the format 'entity, category'"
    )
    parser_categorize.add_argument(
        "--results", type=str, required=True,
        help="Output CSV file."
    )
    parser_categorize.add_argument(
        "--relationships", type=str, nargs="+",
        default=DEFAULT_ANCESTOR_RELATIONSHIPS,
        help="Relationship types that define an ancestor"
    )
    parser_categorize.set_defaults(func=lambda args: categorize_from_file(
        entity_filename=args.entities,
        results_filename=args.results,
        category_filename=args.categories,
        entity_synonyms_filename=args.entity_synonyms,
        category_synonyms_filename=args.category_synonyms,
        manual_categories_filename=args.manual_categories,
        relationships=args.relationships,
    ))

    # -------------------------------------------------------------------------
    # Parse and run
    # -------------------------------------------------------------------------
    cmdargs = parser.parse_args()

    # Logging
    main_only_quicksetup_rootlogger(level=logging.DEBUG if cmdargs.verbose
                                    else logging.INFO)
    log.debug(f"ChEBI lookup from cardinal_pythonlib=={VERSION_STRING}")

    # Caching
    log.debug(f"Using cache path: {cmdargs.cachepath}")
    set_download_cache_path(cmdargs.cachepath)

    # Do something useful
    cmdargs.func(cmdargs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# cardinal_pythonlib/django/request_cache.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Implement a request cache for Django.**

"""

# https://stackoverflow.com/questions/3151469/per-request-cache-in-django

from threading import currentThread

# noinspection PyUnresolvedReferences
from django.core.cache.backends.locmem import LocMemCache

_request_cache = {}
_installed_middleware = False


def get_request_cache():
    """
    Returns the Django request cache for the current thread.
    Requires that ``RequestCacheMiddleware`` is loaded.
    """
    assert _installed_middleware, 'RequestCacheMiddleware not loaded'
    return _request_cache[currentThread()]


# LocMemCache is a threadsafe local memory cache
class RequestCache(LocMemCache):
    """
    Local memory request cache for Django.
    """
    def __init__(self):
        name = f'locmemcache@{hash(currentThread()):d}'
        params = dict()
        super(RequestCache, self).__init__(name, params)


class RequestCacheMiddleware(object):
    """
    Django middleware to implement a request cache.
    """
    def __init__(self):
        global _installed_middleware
        _installed_middleware = True

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def process_request(self, request):
        cache = _request_cache.get(currentThread()) or RequestCache()
        _request_cache[currentThread()] = cache

        cache.clear()



#!/usr/bin/env python
# cardinal_pythonlib/django/forms.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Additional Django form types and associated cleaners/validators.**

"""

from typing import List

# noinspection PyUnresolvedReferences
from django import forms

from cardinal_pythonlib.nhs import is_valid_nhs_number


# =============================================================================
# Multiple values from a text area
# =============================================================================

def clean_int(x) -> int:
    """
    Returns its parameter as an integer, or raises
    ``django.forms.ValidationError``.
    """
    try:
        return int(x)
    except ValueError:
        raise forms.ValidationError(f"Cannot convert to integer: {x!r}")


def clean_nhs_number(x) -> int:
    """
    Returns its parameter as a valid integer NHS number, or raises
    ``django.forms.ValidationError``.
    """
    try:
        x = int(x)
        if not is_valid_nhs_number(x):
            raise ValueError
        return x
    except ValueError:
        raise forms.ValidationError(f"Not a valid NHS number: {x!r}")


class MultipleIntAreaField(forms.Field):
    """
    Django ``forms.Field`` to capture multiple integers.
    """
    # See also https://stackoverflow.com/questions/29303902/django-form-with-list-of-integers  # noqa
    widget = forms.Textarea

    def clean(self, value) -> List[int]:
        return [clean_int(x) for x in value.split()]


class MultipleNhsNumberAreaField(forms.Field):
    """
    Django ``forms.Field`` to capture multiple NHS numbers.
    """
    widget = forms.Textarea

    def clean(self, value) -> List[int]:
        return [clean_nhs_number(x) for x in value.split()]


class MultipleWordAreaField(forms.Field):
    """
    Django ``forms.Field`` to capture multiple words.
    """
    widget = forms.Textarea

    def clean(self, value) -> List[str]:
        return value.split()


class SingleNhsNumberField(forms.IntegerField):
    """
    Django ``forms.Field`` to capture a single NHS number.
    """
    def clean(self, value) -> int:
        return clean_nhs_number(value)



#!/usr/bin/env python
# cardinal_pythonlib/modules.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to work with Python modules.**

"""

import importlib
from importlib.machinery import ExtensionFileLoader, EXTENSION_SUFFIXES
import inspect
import os
import os.path
import pkgutil
# noinspection PyUnresolvedReferences
from types import ModuleType
from typing import Dict, List, Union

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Module management
# =============================================================================

def import_submodules(package: Union[str, ModuleType],
                      base_package_for_relative_import: str = None,
                      recursive: bool = True) -> Dict[str, ModuleType]:
    """
    Import all submodules of a module, recursively, including subpackages.

    Args:
        package: package (name or actual module)
        base_package_for_relative_import: path to prepend?
        recursive: import submodules too?

    Returns:
        dict: mapping from full module name to module

    """
    # https://stackoverflow.com/questions/3365740/how-to-import-all-submodules
    if isinstance(package, str):
        package = importlib.import_module(package,
                                          base_package_for_relative_import)
    results = {}
    for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
        full_name = package.__name__ + '.' + name
        log.debug("importing: {}", full_name)
        results[full_name] = importlib.import_module(full_name)
        if recursive and is_pkg:
            results.update(import_submodules(full_name))
    return results

# Note slightly nastier way: e.g.
#   # Task imports: everything in "tasks" directory
#   task_modules = glob.glob(os.path.dirname(__file__) + "/tasks/*.py")
#   task_modules = [os.path.basename(f)[:-3] for f in task_modules]
#   for tm in task_modules:
#       __import__(tm, locals(), globals())


# =============================================================================
# For package developers
# =============================================================================

def is_builtin_module(module: ModuleType) -> bool:
    """
    Is this module a built-in module, like ``os``?
    Method is as per :func:`inspect.getfile`.
    """
    assert inspect.ismodule(module)
    return not hasattr(module, "__file__")


def is_module_a_package(module: ModuleType) -> bool:
    assert inspect.ismodule(module)
    return os.path.basename(inspect.getfile(module)) == "__init__.py"


def is_c_extension(module: ModuleType) -> bool:
    """
    Modified from
    https://stackoverflow.com/questions/20339053/in-python-how-can-one-tell-if-a-module-comes-from-a-c-extension.

    ``True`` only if the passed module is a C extension implemented as a
    dynamically linked shared library specific to the current platform.

    Args:
        module: Previously imported module object to be tested.

    Returns:
        bool: ``True`` only if this module is a C extension.

    Examples:

    .. code-block:: python
    
        from cardinal_pythonlib.modules import is_c_extension
        
        import os
        import _elementtree as et
        import numpy
        import numpy.core.multiarray as numpy_multiarray
        
        is_c_extension(os)  # False
        is_c_extension(numpy)  # False
        is_c_extension(et)  # False on my system (Python 3.5.6). True in the original example.
        is_c_extension(numpy_multiarray)  # True

    """  # noqa
    assert inspect.ismodule(module), f'"{module}" not a module.'

    # If this module was loaded by a PEP 302-compliant CPython-specific loader
    # loading only C extensions, this module is a C extension.
    if isinstance(getattr(module, '__loader__', None), ExtensionFileLoader):
        return True

    # If it's built-in, it's not a C extension.
    if is_builtin_module(module):
        return False

    # Else, fallback to filetype matching heuristics.
    #
    # Absolute path of the file defining this module.
    module_filename = inspect.getfile(module)

    # "."-prefixed filetype of this path if any or the empty string otherwise.
    module_filetype = os.path.splitext(module_filename)[1]

    # This module is only a C extension if this path's filetype is that of a
    # C extension specific to the current platform.
    return module_filetype in EXTENSION_SUFFIXES


def contains_c_extension(module: ModuleType,
                         import_all_submodules: bool = True,
                         include_external_imports: bool = False,
                         seen: List[ModuleType] = None) -> bool:
    """
    Extends :func:`is_c_extension` by asking: is this module, or any of its
    submodules, a C extension?

    Args:
        module: Previously imported module object to be tested.
        import_all_submodules: explicitly import all submodules of this module?
        include_external_imports: check modules in other packages that this
            module imports?
        seen: used internally for recursion (to deal with recursive modules);
            should be ``None`` when called by users

    Returns:
        bool: ``True`` only if this module or one of its submodules is a C
        extension.

    Examples:

    .. code-block:: python

        import logging
        from cardinal_pythonlib.modules import contains_c_extension
        from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger
        
        import _elementtree as et
        import os
        
        import arrow
        import alembic
        import django
        import numpy
        import numpy.core.multiarray as numpy_multiarray
        
        log = logging.getLogger(__name__)
        # logging.basicConfig(level=logging.DEBUG)  # be verbose
        main_only_quicksetup_rootlogger(level=logging.DEBUG)
        
        contains_c_extension(os)  # False
        contains_c_extension(et)  # False
        
        contains_c_extension(numpy)  # True -- different from is_c_extension()
        contains_c_extension(numpy_multiarray)  # True
        
        contains_c_extension(arrow)  # False
        
        contains_c_extension(alembic)  # False
        contains_c_extension(alembic, include_external_imports=True)  # True
        # ... this example shows that Alembic imports hashlib, which can import
        #     _hashlib, which is a C extension; however, that doesn't stop us (for
        #     example) installing Alembic on a machine with no C compiler
        
        contains_c_extension(django)

    """  # noqa
    assert inspect.ismodule(module), f'"{module}" not a module.'

    if seen is None:  # only true for the top-level call
        seen = []  # type: List[ModuleType]
    if module in seen:  # modules can "contain" themselves
        # already inspected; avoid infinite loops
        return False
    seen.append(module)

    # Check the thing we were asked about
    is_c_ext = is_c_extension(module)
    log.info("Is module {!r} a C extension? {}", module, is_c_ext)
    if is_c_ext:
        return True
    if is_builtin_module(module):
        # built-in, therefore we stop searching it
        return False

    # Now check any children, in a couple of ways

    top_level_module = seen[0]
    top_path = os.path.dirname(top_level_module.__file__)

    # Recurse using dir(). This picks up modules that are automatically
    # imported by our top-level model. But it won't pick up all submodules;
    # try e.g. for django.
    for candidate_name in dir(module):
        candidate = getattr(module, candidate_name)
        # noinspection PyBroadException
        try:
            if not inspect.ismodule(candidate):
                # not a module
                continue
        except Exception:
            # e.g. a Django module that won't import until we configure its
            # settings
            log.error("Failed to test ismodule() status of {!r}", candidate)
            continue
        if is_builtin_module(candidate):
            # built-in, therefore we stop searching it
            continue

        candidate_fname = getattr(candidate, "__file__")
        if not include_external_imports:
            if os.path.commonpath([top_path, candidate_fname]) != top_path:
                log.debug("Skipping, not within the top-level module's "
                          "directory: {!r}", candidate)
                continue
        # Recurse:
        if contains_c_extension(
                module=candidate,
                import_all_submodules=False,  # only done at the top level, below  # noqa
                include_external_imports=include_external_imports,
                seen=seen):
            return True

    if import_all_submodules:
        if not is_module_a_package(module):
            log.debug("Top-level module is not a package: {!r}", module)
            return False

        # Otherwise, for things like Django, we need to recurse in a different
        # way to scan everything.
        # See https://stackoverflow.com/questions/3365740/how-to-import-all-submodules.  # noqa
        log.debug("Walking path: {!r}", top_path)
        # noinspection PyBroadException
        try:
            for loader, module_name, is_pkg in pkgutil.walk_packages([top_path]):  # noqa
                if not is_pkg:
                    log.debug("Skipping, not a package: {!r}", module_name)
                    continue
                log.debug("Manually importing: {!r}", module_name)
                # noinspection PyBroadException
                try:
                    candidate = loader.find_module(module_name)\
                        .load_module(module_name)  # noqa
                except Exception:
                    # e.g. Alembic "autogenerate" gives: "ValueError: attempted
                    # relative import beyond top-level package"; or Django
                    # "django.core.exceptions.ImproperlyConfigured"
                    log.error("Package failed to import: {!r}", module_name)
                    continue
                if contains_c_extension(
                        module=candidate,
                        import_all_submodules=False,  # only done at the top level  # noqa
                        include_external_imports=include_external_imports,
                        seen=seen):
                    return True
        except Exception:
            log.error("Unable to walk packages further; no C extensions "
                      "detected so far!")
            raise

    return False



#!/usr/bin/env python
# cardinal_pythonlib/getch.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Offers the getch() and kbhit() functions, and other terminal-related stuff.**

In this module, :func:`getch` and :func:`kbhit` are mapped to their OS-specific
versions.

"""

import atexit
import select
import sys

try:
    import msvcrt  # Windows only
    termios = None
    tty = None
except ImportError:
    msvcrt = None
    import termios  # Unix only
    import tty  # requires termios, so Unix only


# =============================================================================
# Read single character, waiting for it
# =============================================================================
# https://stackoverflow.com/questions/510357/python-read-a-single-character-from-the-user  # noqa
# http://home.wlu.edu/~levys/software/kbhit.py
# ... modified a little

def _getch_windows() -> str:
    """
    Under Windows, wets a single character from standard input. Does not echo
    to the screen.
    """
    # noinspection PyUnresolvedReferences
    return msvcrt.getch().decode('utf-8')


def _getch_unix() -> str:
    """
    Under UNIX, gets a single character from standard input. Does not echo to
    the screen. Note that the terminal will have been pre-configured, below.
    """
    return sys.stdin.read(1)


# =============================================================================
# Is a keystroke available?
# =============================================================================
# http://code.activestate.com/recipes/572182-how-to-implement-kbhit-on-linux/
# https://stackoverflow.com/questions/2408560/python-nonblocking-console-input

def _kbhit_windows() -> bool:
    """
    Under Windows: is a keystroke available?
    """
    # noinspection PyUnresolvedReferences
    return msvcrt.kbhit()


def _kbhit_unix() -> bool:
    """
    Under UNIX: is a keystroke available?
    """
    dr, dw, de = select.select([sys.stdin], [], [], 0)
    return dr != []


# =============================================================================
# Configure terminal (UNIX)
# =============================================================================

def set_normal_term() -> None:
    """
    Under UNIX: switch to a normal terminal. (Compare :func:`set_curses_term`.)
    """
    termios.tcsetattr(_fd, termios.TCSAFLUSH, _old_term)


def set_curses_term() -> None:
    """
    Under UNIX: switch to an unbuffered, curses-style terminal. (Compare
    :func:`set_normal_term`.)
    """
    termios.tcsetattr(_fd, termios.TCSAFLUSH, _new_term)


# =============================================================================
# Set up for specific OS
# =============================================================================

if msvcrt:
    # -------------------------------------------------------------------------
    # Windows
    # -------------------------------------------------------------------------
    getch = _getch_windows
    kbhit = _kbhit_windows
else:
    # -------------------------------------------------------------------------
    # Unix
    # -------------------------------------------------------------------------
    getch = _getch_unix
    kbhit = _kbhit_unix

    # save the terminal settings
    _fd = sys.stdin.fileno()
    _new_term = termios.tcgetattr(_fd)
    _old_term = termios.tcgetattr(_fd)
    # new terminal setting unbuffered
    _new_term[3] = (_new_term[3] & ~termios.ICANON & ~termios.ECHO)
    _new_term[6][termios.VMIN] = 1
    _new_term[6][termios.VTIME] = 0

    atexit.register(set_normal_term)
    set_curses_term()



#!/usr/bin/env python
# cardinal_pythonlib/httpconst.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Constants for use with HTTP.**

Many of these can be extracted:

.. code-block:: python

    import mimetypes
    mimetypes.types_map['.zip']  # application/zip -- this is built in
    mimetypes.types_map['.xlsx']  # fails
    mimetypes.init()
    mimetypes.types_map['.xlsx']  # application/vnd.openxmlformats-officedocument.spreadsheetml.sheet
    # ... must read some local thing...

Something's changed -- in Python 3.6.8, there's no need for the init() call.
There is also a guessing function, :func:`mimetypes.guess_type`; see
https://docs.python.org/3.6/library/mimetypes.html.

.. code-block:: python

    >>> import mimetypes
    >>> print(mimetypes.guess_type("thing.html"))
    ('text/html', None)
    >>> print(mimetypes.guess_type("thing.xls"))
    ('application/vnd.ms-excel', None)
    >>> print(mimetypes.guess_type("thing.xlsx"))
    ('application/vnd.openxmlformats-officedocument.spreadsheetml.sheet', None)

"""  # noqa


# =============================================================================
# HTTP methods
# =============================================================================

class HttpMethod(object):
    """
    HTTP request methods, as upper-case constants.

    - https://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html
    - https://datatracker.ietf.org/doc/html/rfc7231
    - https://en.wikipedia.org/wiki/Hypertext_Transfer_Protocol#Request_methods
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
    """
    CONNECT = "CONNECT"  # HTTP/1.1
    DELETE = "DELETE"  # HTTP/1.1
    GET = "GET"  # HTTP/1.0
    HEAD = "HEAD"  # HTTP/1.0
    OPTIONS = "OPTIONS"  # HTTP/1.1
    PATCH = "PATCH"  # HTTP/1.1
    POST = "POST"  # HTTP/1.0
    PUT = "PUT"  # HTTP/1.1
    TRACE = "TRACE"  # HTTP/1.1


# =============================================================================
# HTTP status codes
# =============================================================================

class HttpStatus(object):
    """
    HTTP status codes.

    https://en.wikipedia.org/wiki/List_of_HTTP_status_codes
    """
    # 1xx: informational
    CONTINUE = 100
    SWITCHING_PROTOCOLS = 101
    PROCESSING = 102
    EARLY_HINTS = 103

    # 2xx: success
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NON_AUTHORITATIVE_INFORMATION = 203
    NO_CONTENT = 204
    RESET_CONTENT = 205
    PARTIAL_CONTENT = 206
    MULTI_STATUS = 207
    ALREADY_REPORTED = 208
    IM_USED = 226

    # 3xx: redirection
    MULTIPLE_CHOICES = 300
    MOVED_PERMANENTLY = 301
    FOUND = 302
    SEE_OTHER = 303
    NOT_MODIFIED = 304
    USE_PROXY = 305
    SWITCH_PROXY = 306  # no longer used
    TEMPORARY_REDIRECT = 307
    PERMANENT_REDIRECT = 308

    # 4xx: client error
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    PROXY_AUTHENTICATION_REQUIRED = 407
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    GONE = 410
    LENGTH_REQUIRED = 411
    PRECONDTION_FAILED = 412
    PAYLOAD_TOO_LARGE = 413
    URI_TOO_LONG = 414
    UNSUPPORTED_MEDIA_TYPE = 415
    RANGE_NOT_SUITABLE = 416
    EXPECTATION_FAILED = 417
    IM_A_TEAPOT = 418
    MISDIRECTED_REQUEST = 421
    UNPROCESSABLE_ENTITY = 422
    LOCKED = 423
    FAILED_DEPENDENCY = 424
    TOO_EARLY = 425
    UPGRADE_REQUIRED = 426
    PRECONDITION_REQUIRED = 428
    TOO_MANY_REQUESTS = 429
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431
    UNAVAILABLE_FOR_LEGAL_REASONS = 451

    # 5xx: server error
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504
    HTTP_VERSION_NOT_SUPPORTED = 505
    VARIANT_ALSO_NEGOTIATES = 506
    INSUFICIENT_STORAGE = 507
    LOOP_DETECTED = 508
    NOT_EXTENDED = 510
    NETWORK_AUTHENTICATION_REQUIRED = 511

    @classmethod
    def is_good_answer(cls, status: int) -> bool:
        """
        Is the given HTTP status code a satisfactory (happy) answer to a
        client's request?
        """
        return 200 <= status <= 299 or status == cls.PROCESSING


# =============================================================================
# MIME types
# =============================================================================

class MimeType(object):
    """
    Some MIME type constants.
    See also the Python standard library ``mimetypes``; e.g.

    .. code-block:: python

        import mimetypes
        mimetypes.types_map['.pdf']  # 'application/pdf'

    See:

    - Binary:

      - https://stackoverflow.com/questions/6783921/which-mime-type-to-use-for-a-binary-file-thats-specific-to-my-program

    - CSV

      - https://stackoverflow.com/questions/264256/what-is-the-best-mime-type-and-extension-to-use-when-exporting-tab-delimited
      - https://www.iana.org/assignments/media-types/text/tab-separated-values

    - JSON:

      - https://stackoverflow.com/questions/477816/what-is-the-correct-json-content-type

    - ZIP

      - https://stackoverflow.com/questions/4411757/zip-mime-types-when-to-pick-which-one

    - Microsoft Office

      - https://filext.com/faq/office_mime_types.html

    - OpenOffice

      - https://www.openoffice.org/framework/documentation/mimetypes/mimetypes.html
      - https://stackoverflow.com/questions/31489757/what-is-correct-mimetype-with-apache-openoffice-files-like-odt-ods-odp

    """  # noqa
    BINARY = "application/octet-stream"
    CSV = "text/csv"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"  # noqa
    DOT = DOC
    DOTX = "application/vnd.openxmlformats-officedocument.wordprocessingml.template"  # noqa
    FORCE_DOWNLOAD = "application/force-download"
    HTML = "text/html"
    JSON = "application/json"
    ODP = "application/vnd.oasis.opendocument.presentation"
    ODS = "application/vnd.oasis.opendocument.spreadsheet"
    ODT = "application/vnd.oasis.opendocument.text"
    PDF = "application/pdf"
    PNG = "image/png"
    PPT = "application/vnd.ms-powerpoint"
    SQLITE3 = "application/x-sqlite3"
    TEXT = "text/plain"
    TSV = "text/tab-separated-values"
    TXT = TEXT
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    XML = "text/xml"
    ZIP = "application/zip"


ContentType = MimeType



#!/usr/bin/env python
# cardinal_pythonlib/platformfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support for building software.**

"""

import io
import os
import platform
import subprocess
import sys
from typing import Callable, Dict, List, TextIO, Tuple

from cardinal_pythonlib.cmdline import cmdline_quote
from cardinal_pythonlib.fileops import (
    mkdir_p,
    pushd,
    require_executable,
    which_and_require,
)
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.network import download
from cardinal_pythonlib.tee import teed_call

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Types
# =============================================================================

RunFuncType = Callable
# ... we were using Callable[[List[str]], Any] but that caused type-checking
#     errors with functions that also took keyword arguments
# ... i.e. something that looks like:
#     def somefunc(strlist, **kwargs) -> ...
# ... you can't represent that exactly with Callable;
#     https://docs.python.org/3/library/typing.html#typing.Callable
# ... so the best is Callable, which is Callable[..., Any]


# =============================================================================
# Download things
# =============================================================================

def download_if_not_exists(url: str, filename: str,
                           skip_cert_verify: bool = True,
                           mkdir: bool = True) -> None:
    """
    Downloads a URL to a file, unless the file already exists.
    """
    if os.path.isfile(filename):
        log.info("No need to download, already have: {}", filename)
        return
    if mkdir:
        directory, basename = os.path.split(os.path.abspath(filename))
        mkdir_p(directory)
    download(url=url,
             filename=filename,
             skip_cert_verify=skip_cert_verify)


# =============================================================================
# Git functions
# =============================================================================

def git_clone(prettyname: str, url: str, directory: str,
              branch: str = None,
              commit: str = None,
              clone_options: List[str] = None,
              run_func: RunFuncType = None,
              git_executable: str = None) -> bool:
    """
    Fetches a Git repository, unless we have it already.

    Args:
        prettyname: name to display to user
        url: URL
        directory: destination directory
        branch: repository branch
        commit: repository commit tag
        clone_options: additional options to pass to ``git clone``
        run_func: function to use to call an external command
        git_executable: name of git executable (default ``git``)

    Returns:
        did we need to do anything?
    """
    git = which_and_require(git_executable or "git")
    run_func = run_func or subprocess.check_call
    clone_options = clone_options or []  # type: List[str]
    if os.path.isdir(directory):
        log.info(
            "Not re-cloning {} Git repository: using existing source in {}",
            prettyname, directory)
        return False
    log.info("Fetching {} source from {} into {}",
             prettyname, url, directory)
    gitargs = [git, "clone"] + clone_options
    if branch:
        gitargs += ["--branch", branch]
    gitargs += [url, directory]
    run_func(gitargs)
    if commit:
        log.info("Resetting {} local Git repository to commit {}",
                 prettyname, commit)
        run_func([git,
                  "-C", directory,
                  "reset", "--hard", commit])
        # Using a Git repository that's not in the working directory:
        # https://stackoverflow.com/questions/1386291/git-git-dir-not-working-as-expected  # noqa
    return True


# def fix_git_repo_for_windows(directory: str):
#     # https://github.com/openssl/openssl/issues/174
#     log.info("Fixing repository {!r} for Windows line endings", directory)
#     with pushd(directory):
#         run([GIT, "config", "--local", "core.autocrlf", "false"])
#         run([GIT, "config", "--local", "core.eol", "lf"])
#         run([GIT, "rm", "--cached", "-r", "."])
#         run([GIT, "reset", "--hard"])


# =============================================================================
# tar functions
# =============================================================================

def tar_supports_force_local_switch(tar_executable: str) -> bool:
    """
    Does ``tar`` support the ``--force-local`` switch? We ask it.
    """
    tarhelp = fetch([tar_executable, "--help"])
    return "--force-local" in tarhelp


def untar_to_directory(tarfile: str,
                       directory: str,
                       verbose: bool = False,
                       gzipped: bool = False,
                       skip_if_dir_exists: bool = True,
                       run_func: RunFuncType = None,
                       chdir_via_python: bool = True,
                       tar_executable: str = None,
                       tar_supports_force_local: bool = None) -> None:
    """
    Unpacks a TAR file into a specified directory.

    Args:
        tarfile:
            filename of the ``.tar`` file
        directory:
            destination directory
        verbose:
            be verbose?
        gzipped:
            is the ``.tar`` also gzipped, e.g. a ``.tar.gz`` file?
        skip_if_dir_exists:
            don't do anything if the destrination directory exists?
        run_func:
            function to use to call an external command
        chdir_via_python:
            change directory via Python, not via ``tar``. Consider using this
            via Windows, because Cygwin ``tar`` v1.29 falls over when given a
            Windows path for its ``-C`` (or ``--directory``) option.
        tar_executable:
            name of the ``tar`` executable (default is ``tar``)
        tar_supports_force_local:
            does tar support the ``--force-local`` switch? If you pass ``None``
            (the default), this is checked directly via ``tar --help``.
            Linux/GNU tar does; MacOS tar doesn't; Cygwin tar does; Windows 10
            (build 17063+) tar doesn't.
    """
    if skip_if_dir_exists and os.path.isdir(directory):
        log.info("Skipping extraction of {} as directory {} exists",
                 tarfile, directory)
        return
    tar = which_and_require(tar_executable or "tar")
    if tar_supports_force_local is None:
        tar_supports_force_local = tar_supports_force_local_switch(tar)
    log.info("Extracting {} -> {}", tarfile, directory)
    mkdir_p(directory)
    args = [tar, "-x"]  # -x: extract
    if verbose:
        args.append("-v")  # -v: verbose
    if gzipped:
        args.append("-z")  # -z: decompress using gzip
    if tar_supports_force_local:
        args.append("--force-local")  # allows filenames with colons in
    args.extend(["-f", tarfile])  # -f: filename follows
    if chdir_via_python:
        with pushd(directory):
            run_func(args)
    else:
        # chdir via tar
        args.extend(["-C", directory])  # -C: change to directory
        run_func(args)


# =============================================================================
# Environment functions
# =============================================================================

def make_copy_paste_env(env: Dict[str, str]) -> str:
    """
    Convert an environment into a set of commands that can be copied/pasted, on
    the build platform, to recreate that environment.
    """
    windows = platform.system() == "Windows"
    cmd = "set" if windows else "export"
    return (
        "\n".join(
            "{cmd} {k}={v}".format(
                cmd=cmd,
                k=k,
                v=env[k] if windows else subprocess.list2cmdline([env[k]])
            ) for k in sorted(env.keys())
        )
    )
    # Note that even subprocess.list2cmdline() will put needless quotes in
    # here, whereas SET is happy with e.g. SET x=C:\Program Files\somewhere;
    # subprocess.list2cmdline() will also mess up trailing backslashes (e.g.
    # for the VS140COMNTOOLS environment variable).


# =============================================================================
# Run subprocesses in a very verbose way
# =============================================================================

def run(args: List[str],
        env: Dict[str, str] = None,
        capture_stdout: bool = False,
        echo_stdout: bool = True,
        capture_stderr: bool = False,
        echo_stderr: bool = True,
        debug_show_env: bool = True,
        encoding: str = sys.getdefaultencoding(),
        allow_failure: bool = False,
        **kwargs) -> Tuple[str, str]:
    """
    Runs an external process, announcing it.

    Optionally, retrieves its ``stdout`` and/or ``stderr`` output (if not
    retrieved, the output will be visible to the user).

    Args:
        args: list of command-line arguments (the first being the executable)

        env: operating system environment to use (if ``None``, the current OS
            environment will be used)

        capture_stdout: capture the command's ``stdout``?

        echo_stdout: allow the command's ``stdout`` to go to ``sys.stdout``?

        capture_stderr: capture the command's ``stderr``?

        echo_stderr: allow the command's ``stderr`` to go to ``sys.stderr``?

        debug_show_env: be verbose and show the environment used before calling

        encoding: encoding to use to translate the command's output

        allow_failure: if ``True``, continues if the command returns a
            non-zero (failure) exit code; if ``False``, raises an error if
            that happens

        kwargs: additional arguments to :func:`teed_call`

    Returns:
        a tuple: ``(stdout, stderr)``. If the output wasn't captured, an empty
        string will take its place in this tuple.
    """
    cwd = os.getcwd()
    # log.debug("External command Python form: {}", args)
    copy_paste_cmd = cmdline_quote(args)
    csep = "=" * 79
    esep = "-" * 79
    effective_env = env if env is not None else os.environ
    if debug_show_env:
        log.debug(
            "Environment for the command that follows:\n"
            "{esep}\n"
            "{env}\n"
            "{esep}",
            esep=esep,
            env=make_copy_paste_env(effective_env)
        )
    log.info(
        "Launching external command:\n"
        "{csep}\n"
        "WORKING DIRECTORY: {cwd}\n"
        "PYTHON ARGS: {pyargs!r}\n"
        "COMMAND: {cmd}\n"
        "{csep}",
        csep=csep,
        cwd=cwd,
        cmd=copy_paste_cmd,
        pyargs=args
    )
    try:
        with io.StringIO() as out, io.StringIO() as err:
            stdout_targets = []  # type: List[TextIO]
            stderr_targets = []  # type: List[TextIO]
            if capture_stdout:
                stdout_targets.append(out)
            if echo_stdout:
                stdout_targets.append(sys.stdout)
            if capture_stderr:
                stderr_targets.append(err)
            if echo_stderr:
                stderr_targets.append(sys.stderr)
            retcode = teed_call(args,
                                stdout_targets=stdout_targets,
                                stderr_targets=stderr_targets,
                                encoding=encoding,
                                env=env,
                                **kwargs)
            stdout = out.getvalue()
            stderr = err.getvalue()
            if retcode != 0 and not allow_failure:
                # subprocess.check_call() and check_output() raise
                # CalledProcessError if the called process returns a non-zero
                # return code.
                raise subprocess.CalledProcessError(returncode=retcode,
                                                    cmd=args,
                                                    output=stdout,
                                                    stderr=stderr)
        log.debug("\n{csep}\nFINISHED SUCCESSFULLY: {cmd}\n{csep}",
                  cmd=copy_paste_cmd, csep=csep)
        return stdout, stderr
    except FileNotFoundError:
        require_executable(args[0])  # which is missing, so we'll see some help
        raise
    except subprocess.CalledProcessError:
        log.critical(
            "Command that failed:\n"
            "[ENVIRONMENT]\n"
            "{env}\n"
            "\n"
            "[DIRECTORY] {cwd}\n"
            "[PYTHON ARGS] {pyargs}\n"
            "[COMMAND] {cmd}",
            cwd=cwd,
            env=make_copy_paste_env(effective_env),
            cmd=copy_paste_cmd,
            pyargs=args
        )
        raise


def fetch(args: List[str], env: Dict[str, str] = None,
          encoding: str = sys.getdefaultencoding()) -> str:
    """
    Run a command and returns its stdout.

    Args:
        args: the command-line arguments
        env: the operating system environment to use
        encoding: the encoding to use for ``stdout``

    Returns:
        the command's ``stdout`` output

    """
    stdout, _ = run(args, env=env, capture_stdout=True,
                    echo_stdout=False, encoding=encoding)
    log.debug("{}", stdout)
    return stdout



#!/usr/bin/env python
# cardinal_pythonlib/errorreporter_mw.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**WSGI middleware to produce tracebacks upon error.**

"""

# =============================================================================
# ErrorReportingMiddleware
# =============================================================================
# From: http://pylonsbook.com/en/1.0/the-web-server-gateway-interface-wsgi.html
# Modified to use six.StringIO
# Latest changes: 6 Jan 2016

import cgitb
# import six
from io import StringIO
import sys
from typing import List


from cardinal_pythonlib.wsgi.constants import (
    TYPE_WSGI_APP,
    TYPE_WSGI_APP_RESULT,
    TYPE_WSGI_ENVIRON,
    TYPE_WSGI_EXC_INFO,
    TYPE_WSGI_START_RESPONSE,
)


class ErrorReportingMiddleware(object):
    """
    WSGI middleware to produce ``cgitb`` traceback upon errors.
    """
    def __init__(self, app: TYPE_WSGI_APP) -> None:
        self.app = app

    @staticmethod
    def format_exception(exc_info: TYPE_WSGI_EXC_INFO) -> List[bytes]:
        dummy_file = StringIO()
        hook = cgitb.Hook(file=dummy_file)
        hook(*exc_info)
        return [dummy_file.getvalue().encode('utf-8')]

    def __call__(self,
                 environ: TYPE_WSGI_ENVIRON,
                 start_response: TYPE_WSGI_START_RESPONSE) \
            -> TYPE_WSGI_APP_RESULT:
        # noinspection PyBroadException
        try:
            return self.app(environ, start_response)
        except Exception:
            exc_info = sys.exc_info()
            start_response(
                '500 Internal Server Error',
                [('content-type', 'text/html; charset=utf-8')],
                exc_info
            )
            return self.format_exception(exc_info)



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/session.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to work with SQLAlchemy sessions/engines.**

"""

from typing import TYPE_CHECKING
import os

from sqlalchemy.engine import create_engine
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm.session import Session

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Database URLs
# =============================================================================

SQLITE_MEMORY_URL = "sqlite://"


def make_mysql_url(username: str, password: str, dbname: str,
                   driver: str = "mysqldb", host: str = "localhost",
                   port: int = 3306, charset: str = "utf8") -> str:
    """
    Makes an SQLAlchemy URL for a MySQL database.
    """
    return "mysql+{driver}://{u}:{p}@{host}:{port}/{db}?charset={cs}".format(
        driver=driver,
        host=host,
        port=port,
        db=dbname,
        u=username,
        p=password,
        cs=charset,
    )


def make_sqlite_url(filename: str) -> str:
    """
    Makes an SQLAlchemy URL for a SQLite database.
    """
    absfile = os.path.abspath(filename)
    return "sqlite://{host}/{path}".format(host="", path=absfile)
    # ... makes it clear how it works! Ends up being sqlite:////abspath
    # or sqlite:///relpath. Also works with backslashes for Windows paths; see
    # http://docs.sqlalchemy.org/en/latest/core/engines.html#sqlite


# =============================================================================
# Connection management
# =============================================================================

def get_engine_from_session(dbsession: Session) -> Engine:
    """
    Gets the SQLAlchemy :class:`Engine` from a SQLAlchemy :class:`Session`.
    """
    engine = dbsession.bind
    assert isinstance(engine, Engine)
    return engine


def get_safe_url_from_engine(engine: Engine) -> str:
    """
    Gets a URL from an :class:`Engine`, obscuring the password.
    """
    raw_url = engine.url  # type: str
    url_obj = make_url(raw_url)  # type: URL
    return repr(url_obj)
    # The default repr() implementation calls
    # self.__to_string__(hide_password=False)


def get_safe_url_from_session(dbsession: Session) -> str:
    """
    Gets a URL from a :class:`Session`, obscuring the password.
    """
    return get_safe_url_from_engine(get_engine_from_session(dbsession))


def get_safe_url_from_url(url: str) -> str:
    """
    Converts an SQLAlchemy URL into a safe version that obscures the password.
    """
    engine = create_engine(url)
    return get_safe_url_from_engine(engine)



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/alembic_ops.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for Alembic, specifically to support view creation.**

From https://alembic.readthedocs.org/en/latest/cookbook.html.

HAS NO TYPE ANNOTATIONS - Alembic uses ``inspect.getargspec()``, which chokes
on them.
"""

from alembic.operations import Operations, MigrateOperation


# =============================================================================
# The thing (e.g. view) we are representing
# =============================================================================

class ReplaceableObject(object):
    def __init__(self, name, sqltext):
        """
        Object that can be passed to the ``create_view()`` and similar
        functions that we will register within the ``alembic.op`` namespace.

        See https://alembic.zzzcomputing.com/en/latest/cookbook.html#replaceable-objects

        Args:
            name:
                e.g. name of a view, such as ``subject_session_view``

            sqltext:
                e.g. SQL to create the view, such as

                .. code-block:: sql

                    SELECT
                        C.subject,
                        S.*
                    FROM
                        config C
                        INNER JOIN session S ON S.config_id = C.config_id

        """  # noqa
        self.name = name
        self.sqltext = sqltext


# =============================================================================
# An operation that can be reversed
# =============================================================================

class ReversibleOp(MigrateOperation):
    """
    Represents a DDL (SQL) migration operation that can be reversed; e.g. the
    combination of ``CREATE VIEW`` and ``DROP VIEW``.
    """
    def __init__(self, target):
        """
        Args:
            target: instance of :class:`.ReplaceableObject`
        """
        self.target = target

    @classmethod
    def invoke_for_target(cls, operations, target):
        """
        Invokes the operation.

        Args:
            operations: instance of ``alembic.operations.base.Operations``
            target: instance of :class:`.ReplaceableObject`

        Returns:
            result of ``alembic.operations.base.Operations.invoke``
            performed upon an instance of this class initialized with
            ``target``
        """
        op = cls(target)
        return operations.invoke(op)

    def reverse(self):
        """
        Returns:
            the ``MigrateOperation`` representing the reverse of this operation
        """
        raise NotImplementedError()

    @classmethod
    def _get_object_from_version(cls, operations, ident):
        """
        Returns a Python object from an Alembic migration module (script).

        Args:
            operations: instance of ``alembic.operations.base.Operations``
            ident: string of the format ``version.objname``

        Returns:
            the object whose name is ``objname`` within the Alembic migration
            script identified by ``version``
        """
        version, objname = ident.split(".")

        module_ = operations.get_context().script.get_revision(version).module
        obj = getattr(module_, objname)
        return obj

    @classmethod
    def replace(cls, operations, target, replaces=None, replace_with=None):

        if replaces:
            old_obj = cls._get_object_from_version(operations, replaces)
            drop_old = cls(old_obj).reverse()
            create_new = cls(target)
        elif replace_with:
            old_obj = cls._get_object_from_version(operations, replace_with)
            drop_old = cls(target).reverse()
            create_new = cls(old_obj)
        else:
            raise TypeError("replaces or replace_with is required")

        operations.invoke(drop_old)
        operations.invoke(create_new)


# =============================================================================
# Operations that will become part of the op.* namespace
# =============================================================================

@Operations.register_operation("create_view", "invoke_for_target")
@Operations.register_operation("replace_view", "replace")
class CreateViewOp(ReversibleOp):
    """
    Represents ``CREATE VIEW`` (reversed by ``DROP VIEW``).
    """
    def reverse(self):
        return DropViewOp(self.target)


@Operations.register_operation("drop_view", "invoke_for_target")
class DropViewOp(ReversibleOp):
    """
    Represents ``DROP VIEW`` (reversed by ``CREATE VIEW``).
    """
    def reverse(self):
        return CreateViewOp(self.view)


@Operations.register_operation("create_sp", "invoke_for_target")
@Operations.register_operation("replace_sp", "replace")
class CreateSPOp(ReversibleOp):
    """
    Represents ``CREATE FUNCTION`` (reversed by ``DROP FUNCTION``)
    [sp = stored procedure].
    """
    def reverse(self):
        return DropSPOp(self.target)


@Operations.register_operation("drop_sp", "invoke_for_target")
class DropSPOp(ReversibleOp):
    """
    Represents ``DROP FUNCTION`` (reversed by ``CREATE FUNCTION``)
    [sp = stored procedure].
    """
    def reverse(self):
        return CreateSPOp(self.target)


# =============================================================================
# Implementation of these operations
# =============================================================================

@Operations.implementation_for(CreateViewOp)
def create_view(operations, operation):
    """
    Implements ``CREATE VIEW``.

    Args:
        operations: instance of ``alembic.operations.base.Operations``
        operation: instance of :class:`.ReversibleOp`

    Returns:
        ``None``
    """
    operations.execute("CREATE VIEW %s AS %s" % (
        operation.target.name,
        operation.target.sqltext
    ))


@Operations.implementation_for(DropViewOp)
def drop_view(operations, operation):
    """
    Implements ``DROP VIEW``.

    Args:
        operations: instance of ``alembic.operations.base.Operations``
        operation: instance of :class:`.ReversibleOp`

    Returns:
        ``None``
    """
    operations.execute("DROP VIEW %s" % operation.target.name)


@Operations.implementation_for(CreateSPOp)
def create_sp(operations, operation):
    """
    Implements ``CREATE FUNCTION``.

    Args:
        operations: instance of ``alembic.operations.base.Operations``
        operation: instance of :class:`.ReversibleOp`

    Returns:
        ``None``
    """
    operations.execute(
        "CREATE FUNCTION %s %s" % (
            operation.target.name, operation.target.sqltext
        )
    )


@Operations.implementation_for(DropSPOp)
def drop_sp(operations, operation):
    """
    Implements ``DROP FUNCTION``.

    Args:
        operations: instance of ``alembic.operations.base.Operations``
        operation: instance of :class:`.ReversibleOp`

    Returns:
        ``None``
    """
    operations.execute("DROP FUNCTION %s" % operation.target.name)



#!/usr/bin/env python
# cardinal_pythonlib/django/serve.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Helper functions to serve specific types of content via Django (e.g. raw
files, PDFs).**

"""


import os
from typing import Iterable, Optional, Union

# noinspection PyUnresolvedReferences
from django.conf import settings
# noinspection PyUnresolvedReferences
from django.http import FileResponse, HttpResponse
# noinspection PyUnresolvedReferences
from django.http.response import HttpResponseBase
# noinspection PyUnresolvedReferences
from django.utils.encoding import smart_str

from cardinal_pythonlib.httpconst import MimeType
from cardinal_pythonlib.pdf import (
    get_concatenated_pdf_from_disk,
    get_concatenated_pdf_in_memory,
    get_pdf_from_html,
    PdfPlan,
)


# =============================================================================
# File serving
# =============================================================================

# I thought this function was superseded by django-filetransfers:
# http://nemesisdesign.net/blog/coding/django-private-file-upload-and-serving/
# https://www.allbuttonspressed.com/projects/django-filetransfers
# ... but it turns out that filetransfers.api.serve_file uses a file object,
# not a filename. Not impossible, but never mind.

def add_http_headers_for_attachment(
        response: HttpResponse,
        offered_filename: str = None,
        content_type: str = None,
        as_attachment: bool = False,
        as_inline: bool = False,
        content_length: int = None,
        default_content_type: Optional[str] = MimeType.FORCE_DOWNLOAD) -> None:
    """
    Add HTTP headers to a Django response class object.

    Args:

        response:
            ``HttpResponse`` instance
        offered_filename:
            filename that the client browser will suggest
        content_type:
            HTTP content type
        as_attachment:
            If True, browsers will generally save to disk.
            If False, they may display it inline.
            https://www.w3.org/Protocols/rfc2616/rfc2616-sec19.html
        as_inline:
            attempt to force inline (only if not as_attachment)
        content_length:
            HTTP content length
        default_content_type:
            HTTP content type to use as default, if ``content_type`` is
            ``None``

    """
    # Parameters
    if offered_filename is None:
        offered_filename = ''
    if content_type is None:
        content_type = default_content_type

    # Content-Type
    if content_type is not None:
        response['Content-Type'] = content_type

    # Content-Disposition
    if as_attachment:
        prefix = 'attachment; '
    elif as_inline:
        prefix = 'inline; '
    else:
        prefix = ''
    fname = 'filename=%s' % smart_str(offered_filename)
    response['Content-Disposition'] = prefix + fname

    # Content-Length
    if content_length is not None:
        response['Content-Length'] = content_length


def serve_file(path_to_file: str,
               offered_filename: str = None,
               content_type: str = None,
               as_attachment: bool = False,
               as_inline: bool = False,
               default_content_type: Optional[str] = MimeType.FORCE_DOWNLOAD) \
        -> HttpResponseBase:
    """
    Serve up a file from disk.

    There are two methods. One is chosen via ``settings.XSENDFILE`` (a
    non-standard value in the Django settings file, defaulting to ``False``):

    (a) serve directly (if ``XSENDFILE`` is absent or False);
    (b) serve by asking the web server to do so via the X-SendFile directive
        (if ``XSENDFILE`` is True).

    Args:

        path_to_file:
            path to file on our server
        offered_filename:
            filename that the client browser will suggest
        content_type:
            HTTP content type
        as_attachment:
            If True, browsers will generally save to disk.
            If False, they may display it inline.
            https://www.w3.org/Protocols/rfc2616/rfc2616-sec19.html
        as_inline:
            attempt to force inline (only if not as_attachment)
        default_content_type:
            HTTP content type to use as default, if ``content_type`` is
            ``None``
    """
    # https://stackoverflow.com/questions/1156246/having-django-serve-downloadable-files  # noqa
    # https://docs.djangoproject.com/en/dev/ref/request-response/#telling-the-browser-to-treat-the-response-as-a-file-attachment  # noqa
    # https://djangosnippets.org/snippets/365/
    if offered_filename is None:
        offered_filename = os.path.basename(path_to_file) or ''
    if getattr(settings, "XSENDFILE", False):
        response = HttpResponse()
        response['X-Sendfile'] = smart_str(path_to_file)
        content_length = os.path.getsize(path_to_file)
    else:
        response = FileResponse(open(path_to_file, mode='rb'))
        content_length = None
    add_http_headers_for_attachment(
        response,
        offered_filename=offered_filename,
        content_type=content_type,
        as_attachment=as_attachment,
        as_inline=as_inline,
        content_length=content_length,
        default_content_type=default_content_type)
    return response
    # Note for debugging: Chrome may request a file more than once (e.g. with a
    # GET request that's then marked 'canceled' in the Network tab of the
    # developer console); this is normal:
    #   https://stackoverflow.com/questions/4460661/what-to-do-with-chrome-sending-extra-requests  # noqa


def serve_buffer(
        data: bytes,
        offered_filename: str = None,
        content_type: str = None,
        as_attachment: bool = True,
        as_inline: bool = False,
        default_content_type: Optional[str] = MimeType.FORCE_DOWNLOAD) \
        -> HttpResponse:
    """
    Serve up binary data from a buffer.
    Options as for :func:`serve_file`.
    """
    response = HttpResponse(data)
    add_http_headers_for_attachment(
        response,
        offered_filename=offered_filename,
        content_type=content_type,
        as_attachment=as_attachment,
        as_inline=as_inline,
        content_length=len(data),
        default_content_type=default_content_type)
    return response


# =============================================================================
# Simpler versions
# =============================================================================

def add_download_filename(response: HttpResponse, filename: str) -> None:
    """
    Adds a ``Content-Disposition`` header to the HTTP response to say that
    there is an attachment with the specified filename.
    """
    # https://docs.djangoproject.com/en/1.9/howto/outputting-csv/
    add_http_headers_for_attachment(response)
    response['Content-Disposition'] = f'attachment; filename="{filename}"'


def file_response(data: Union[bytes, str],  # HttpResponse encodes str if req'd
                  content_type: str,
                  filename: str) -> HttpResponse:
    """
    Returns an ``HttpResponse`` with an attachment containing the specified
    data with the specified filename as an attachment.
    """
    response = HttpResponse(data, content_type=content_type)
    add_download_filename(response, filename)
    return response


# =============================================================================
# PDF serving
# =============================================================================

def serve_concatenated_pdf_from_disk(
        filenames: Iterable[str],
        offered_filename: str = "crate_download.pdf",
        **kwargs) -> HttpResponse:
    """
    Concatenates PDFs from disk and serves them.
    """
    pdf = get_concatenated_pdf_from_disk(filenames, **kwargs)
    return serve_buffer(pdf,
                        offered_filename=offered_filename,
                        content_type=MimeType.PDF,
                        as_attachment=False,
                        as_inline=True)


def serve_pdf_from_html(html: str,
                        offered_filename: str = "test.pdf",
                        **kwargs) -> HttpResponse:
    """
    Same args as ``pdf_from_html()``.
    WATCH OUT: may not apply e.g. wkhtmltopdf options as you'd wish.
    """
    pdf = get_pdf_from_html(html, **kwargs)
    return serve_buffer(pdf,
                        offered_filename=offered_filename,
                        content_type=MimeType.PDF,
                        as_attachment=False,
                        as_inline=True)


def serve_concatenated_pdf_from_memory(
        pdf_plans: Iterable[PdfPlan],
        start_recto: bool = True,
        offered_filename: str = "crate_download.pdf") -> HttpResponse:
    """
    Concatenates PDFs into memory and serves it.
    WATCH OUT: may not apply e.g. wkhtmltopdf options as you'd wish.
    """
    pdf = get_concatenated_pdf_in_memory(pdf_plans, start_recto=start_recto)
    return serve_buffer(pdf,
                        offered_filename=offered_filename,
                        content_type=MimeType.PDF,
                        as_attachment=False,
                        as_inline=True)



#!/usr/bin/env python
# cardinal_pythonlib/tools/backup_mysql_database.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Command-line tool to back up a MySQL database to disk, via mysqldump.**

"""

import argparse
import datetime
import getpass
import logging
import os
import subprocess
import sys
from typing import List

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


def cmdargs(mysqldump: str,
            username: str,
            password: str,
            database: str,
            verbose: bool,
            with_drop_create_database: bool,
            max_allowed_packet: str,
            hide_password: bool = False) -> List[str]:
    """
    Returns command arguments for a ``mysqldump`` call.

    Args:
        mysqldump: ``mysqldump`` executable filename
        username: user name
        password: password
        database: database name
        verbose: verbose output?
        with_drop_create_database: produce commands to ``DROP`` the database
            and recreate it?
        max_allowed_packet: passed to ``mysqldump``
        hide_password: obscure the password (will break the arguments but
            provide a safe version to show the user)?

    Returns:
        list of command-line arguments
    """
    ca = [
        mysqldump,
        "-u", username,
        "-p{}".format("*****" if hide_password else password),
        f"--max_allowed_packet={max_allowed_packet}",
        "--hex-blob",  # preferable to raw binary in our .sql file
    ]
    if verbose:
        ca.append("--verbose")
    if with_drop_create_database:
        ca.extend([
            "--add-drop-database",
            "--databases",
            database
        ])
    else:
        ca.append(database)
        pass
    return ca


def main() -> None:
    """
    Command-line processor. See ``--help`` for details.
    """
    main_only_quicksetup_rootlogger()
    wdcd_suffix = "_with_drop_create_database"
    timeformat = "%Y%m%dT%H%M%S"
    parser = argparse.ArgumentParser(
        description=f"""
        Back up a specific MySQL database. The resulting filename has the
        format '<DATABASENAME>_<DATETIME><SUFFIX>.sql', where <DATETIME> is of
        the ISO-8601 format {timeformat!r} and <SUFFIX> is either blank or
        {wdcd_suffix!r}.

        A typical filename is therefore 'mydb_20190415T205524.sql'.
        """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "databases", nargs="+",
        help="Database(s) to back up")
    parser.add_argument(
        "--max_allowed_packet", default="1GB",
        help="Maximum size of buffer")
    parser.add_argument(
        "--mysqldump", default="mysqldump",
        help="mysqldump executable")
    parser.add_argument(
        "--username", default="root",
        help="MySQL user")
    parser.add_argument(
        "--password",
        help="MySQL password (AVOID THIS OPTION IF POSSIBLE; VERY INSECURE; "
             "VISIBLE TO OTHER PROCESSES; if you don't use it, you'll be "
             "prompted for the password)")
    parser.add_argument(
        "--with_drop_create_database", action="store_true",
        help="Include DROP DATABASE and CREATE DATABASE commands, and append "
             "a suffix to the output filename as above")
    parser.add_argument(
        "--output_dir", type=str,
        help="Output directory (if not specified, current directory will be "
             "used)")
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output")
    args = parser.parse_args()

    output_dir = args.output_dir or os.getcwd()
    os.chdir(output_dir)

    password = args.password or getpass.getpass(
        prompt=f"MySQL password for user {args.username}: ")

    output_files = []  # type: List[str]
    if args.with_drop_create_database:
        log.info("Note that the DROP DATABASE commands will look like they're "
                 "commented out, but they're not: "
                 "https://dba.stackexchange.com/questions/59294/")
        suffix = wdcd_suffix
    else:
        suffix = ""
    for db in args.databases:
        now = datetime.datetime.now().strftime(timeformat)
        outfilename = f"{db}_{now}{suffix}.sql"
        display_args = cmdargs(
            mysqldump=args.mysqldump,
            username=args.username,
            password=password,
            database=db,
            verbose=args.verbose,
            with_drop_create_database=args.with_drop_create_database,
            max_allowed_packet=args.max_allowed_packet,
            hide_password=True
        )
        actual_args = cmdargs(
            mysqldump=args.mysqldump,
            username=args.username,
            password=password,
            database=db,
            verbose=args.verbose,
            with_drop_create_database=args.with_drop_create_database,
            max_allowed_packet=args.max_allowed_packet,
            hide_password=False
        )
        log.info("Executing: " + repr(display_args))
        log.info("Output file: " + repr(outfilename))
        try:
            with open(outfilename, "w") as f:
                subprocess.check_call(actual_args, stdout=f)
        except subprocess.CalledProcessError:
            os.remove(outfilename)
            log.critical("Failed!")
            sys.exit(1)
        output_files.append(outfilename)
    log.info("Done. See:\n" + "\n".join("    " + x for x in output_files))
    if args.with_drop_create_database:
        log.info("To restore: mysql -u USER -p < BACKUP.sql")
    else:
        log.info("To restore: mysql -u USER -p DATABASE < BACKUP.sql")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/bulk_email/models.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

SQLAlchemy models for the simple bulk e-mail tool.

"""

# =============================================================================
# Imports
# =============================================================================

import datetime
import logging
import sys
from typing import Any, Iterable, Tuple

from sqlalchemy.event import listens_for
from sqlalchemy.orm.attributes import instance_state
from sqlalchemy.orm.exc import NoResultFound
# noinspection PyProtectedMember
from sqlalchemy.orm.session import make_transient, Session
from sqlalchemy.sql.schema import Column, ForeignKey
from sqlalchemy.sql.sqltypes import (
    Boolean,
    DateTime,
    Float,
    Integer,
    String,
    Text,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import and_, exists

from cardinal_pythonlib.bulk_email.constants import (
    CONTENT_TYPE_MAX_LENGTH,
    ENCODING_NAME_MAX_LENGTH,
    FERNET_KEY_BASE64_LENGTH,
    HOSTNAME_MAX_LENGTH,
    PASSWORD_MAX_LENGTH,
    USERNAME_MAX_LENGTH,
)
from cardinal_pythonlib.colander_utils import EMAIL_ADDRESS_MAX_LEN
from cardinal_pythonlib.email.sendmail import (
    ASCII,
    CONTENT_TYPE_TEXT,
    is_email_valid,
    UTF8,
)
from cardinal_pythonlib.sqlalchemy.orm_query import CountStarSpecializedQuery
from cardinal_pythonlib.sysops import EXIT_FAILURE

try:
    from cryptography.fernet import Fernet
except ImportError:
    print("First, please install the 'cryptography' module with "
          "'pip install cryptography'.")
    sys.exit(EXIT_FAILURE)

log = logging.getLogger(__name__)


# =============================================================================
# The master SQLAlchemy ORM class
# =============================================================================

Base = declarative_base()


def make_table_args(*args, **kwargs) -> Tuple[Any]:
    """
    SQLAlchemy allows several formats for __table_args__. The most generic
    is a tuple that is a sequence of e.g. constraints, with the last element
    being a dictionary of keyword arguments.
    """
    # Add standard ones:
    kwargs["mysql_engine"] = "InnoDB"
    kwargs["mysql_row_format"] = "DYNAMIC"
    kwargs["mysql_charset"] = "utf8mb4 COLLATE utf8mb4_unicode_ci"

    # Return the tuple:
    return tuple(args + (kwargs, ))


# =============================================================================
# Config
# =============================================================================

# noinspection PyUnusedLocal
@listens_for(Session, "before_flush")
def before_flush(session: Session, flush_context, instances) -> None:
    # https://docs.sqlalchemy.org/en/14/orm/events.html#sqlalchemy.orm.SessionEvents.before_flush  # noqa
    for instance in session.dirty:
        if not isinstance(instance, Config):
            continue
        if not session.is_modified(instance):
            continue
        if not instance_state(instance).has_identity:
            continue

        # make it transient
        instance.new_version(session)
        # re-add
        session.add(instance)


class Config(Base):
    """
    Stores records of configuration information.

    Maintains history (so the config for previous send jobs is visible), per:

    - https://docs.sqlalchemy.org/en/14/_modules/examples/versioned_rows/versioned_rows.html
    - https://docs.sqlalchemy.org/en/14/orm/examples.html#module-examples.versioned_rows
    """  # noqa

    __tablename__ = "config"
    __table_args__ = make_table_args(
        comment="Stores configuration records."
    )

    # -------------------------------------------------------------------------
    # Columns
    # -------------------------------------------------------------------------

    config_id = Column(
        Integer,
        # ... may not be supported by all databases; see
        # https://docs.sqlalchemy.org/en/14/core/constraints.html#check-constraint  # noqa
        # ... though MySQL has recently added this:
        # https://dev.mysql.com/doc/refman/8.0/en/create-table-check-constraints.html  # noqa
        primary_key=True,
        autoincrement=True,
        comment=f"Primary key."
    )
    database_creation_datetime = Column(
        DateTime,
        comment="Date/time this database was created."
    )
    host = Column(
        String(length=HOSTNAME_MAX_LENGTH),
        nullable=False,
        comment="Mail (SMTP) server hostname"
    )
    port = Column(
        Integer,
        nullable=False,
        comment="Mail (SMTP) server TCP port"
    )
    use_tls = Column(
        Boolean,
        nullable=False,
        comment="Use TLS?"
    )
    username = Column(
        String(length=USERNAME_MAX_LENGTH),
        nullable=False,
        comment="Mail (SMTP) server username"
    )
    password_encrypted = Column(
        String(length=PASSWORD_MAX_LENGTH),
        nullable=False,
        comment="Mail (SMTP) server password (reversibly encrypted)"
    )
    encryption_key = Column(
        String(length=FERNET_KEY_BASE64_LENGTH),
        nullable=False,
        comment="Encryption key for password"
    )
    time_between_emails = Column(
        Float,
        nullable=False,
        comment="Time (in seconds) between each e-mail."
    )

    # -------------------------------------------------------------------------
    # Creation and version control
    # -------------------------------------------------------------------------

    @classmethod
    def get_current_config(cls, session: Session) -> "Config":
        """
        Fetch the current config, or raise.
        """
        cfg = (
            session.query(cls)
            .order_by(cls.config_id.desc())
            .first()  # returns object or None
        )
        if cfg is None:
            error = "No configuration found. " \
                    "Run the 'configure' command first."
            log.critical(error)
            raise ValueError(error)
        return cfg

    @classmethod
    def get_or_create_config(cls, session: Session) -> "Config":
        """
        Fetch the current config, or make a new one.
        """
        cfg = (
            session.query(cls)
            .order_by(cls.config_id.desc())
            .first()  # returns object or None
        )
        if cfg is None:
            cfg = cls()
            session.add(cfg)
        return cfg

    def __init__(self, *args, **kwargs) -> None:
        """
        Default configuration.
        """
        super().__init__(*args, **kwargs)
        self.database_creation_datetime = datetime.datetime.now()

    # noinspection PyUnusedLocal
    def new_version(self, session: Session) -> None:
        """
        Called by listener (see above) when object is being saved.
        """
        # Make us transient (removes persistent identity).
        make_transient(self)

        # Set PK to None. A new PK will be generated on INSERT.
        self.config_id = None

    # -------------------------------------------------------------------------
    # Obscuring passwords
    # -------------------------------------------------------------------------

    @property
    def password(self) -> str:
        """
        Returns the plaintext password.
        """
        # Key
        key_b64_bytes = self.encryption_key.encode(ASCII)
        # Engine
        f = Fernet(key_b64_bytes)
        # Decryption
        encrypted_password_b64_bytes = self.password_encrypted.encode(ASCII)
        return f.decrypt(encrypted_password_b64_bytes).decode(UTF8)

    @password.setter
    def password(self, password: str) -> None:
        """
        Store the password using reversible encryption.
        """
        # Key
        key_b64_bytes = Fernet.generate_key()  # base64-encoded 32-byte key
        self.encryption_key = key_b64_bytes.decode(ASCII)  # str version
        # Engine
        f = Fernet(key_b64_bytes)
        # Encryption
        password_bytes = password.encode(UTF8)
        encrypted_b64_bytes = f.encrypt(password_bytes)  # base64-encoded
        self.password_encrypted = encrypted_b64_bytes.decode(ASCII)


# =============================================================================
# Recipient
# =============================================================================

class Recipient(Base):
    """
    Details of an e-mail recipient.
    """
    __tablename__ = "recipient"
    __table_args__ = make_table_args(
        comment="E-mail recipients."
    )

    # -------------------------------------------------------------------------
    # Columns
    # -------------------------------------------------------------------------

    recipient_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        comment="Primary key."
    )  # type: int
    email = Column(
        String(length=EMAIL_ADDRESS_MAX_LEN),
        nullable=False,
        unique=True,
        comment="E-mail address."
    )

    # -------------------------------------------------------------------------
    # Creation
    # -------------------------------------------------------------------------

    @classmethod
    def fetch_or_add(cls, email: str, session: Session) -> "Recipient":
        """
        Fetch or create/add the recipient.
        """
        try:
            return session.query(cls).filter_by(
                email=email
            ).one()
        except NoResultFound:
            newvar = cls(email=email)
            session.add(newvar)
            return newvar

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not is_email_valid(self.email):
            raise ValueError(f"Invalid e-mail address: {self.email!r}")

    # -------------------------------------------------------------------------
    # Info
    # -------------------------------------------------------------------------

    @staticmethod
    def n_recipients(session: Session) -> int:
        query = CountStarSpecializedQuery([Recipient], session=session)
        return query.count_star()


# =============================================================================
# Content
# =============================================================================

class Content(Base):
    """
    Contents of an e-mail that will be, or has been, sent to one or many
    recipients.
    """
    __tablename__ = "content"
    __table_args__ = make_table_args(
        comment="Contents of an e-mail (without recipient details; these are "
                "templates that may be sent to multiple recipients)."
    )

    # -------------------------------------------------------------------------
    # Columns
    # -------------------------------------------------------------------------

    content_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        comment="Primary key."
    )  # type: int
    from_addr = Column(
        String(length=EMAIL_ADDRESS_MAX_LEN),
        nullable=False,
        comment="From (sender's) e-mail address."
    )
    reply_to_addr = Column(
        String(length=EMAIL_ADDRESS_MAX_LEN),
        nullable=False,
        comment="Reply-To (sender's) e-mail address."
    )
    email_datetime = Column(
        DateTime,
        comment='E-mail creation date/time, or NULL for '
                '"the time of sending".'
    )
    subject = Column(
        Text,
        # There is no limit to the subject length in RFC 2822; you can have
        # multiple lines (each up to 998 characters).
        comment="E-mail subject"
    )
    body = Column(
        Text,
        comment="E-mail body"
    )
    content_type = Column(
        String(length=CONTENT_TYPE_MAX_LENGTH),
        nullable=False,
        default=CONTENT_TYPE_TEXT,
        server_default=CONTENT_TYPE_TEXT,
        comment=f"Character set (encoding). Default: {UTF8!r}."
    )
    charset = Column(
        String(length=ENCODING_NAME_MAX_LENGTH),
        nullable=False,
        default=UTF8,
        server_default=UTF8,
        comment=f"Character set (encoding). Default: {UTF8!r}."
    )

    # -------------------------------------------------------------------------
    # Info
    # -------------------------------------------------------------------------

    @staticmethod
    def n_templates(session: Session) -> int:
        query = CountStarSpecializedQuery([Content], session=session)
        return query.count_star()


# =============================================================================
# Job
# =============================================================================

class Job(Base):
    """
    A task to send one e-mail to one recipient.
    """
    __tablename__ = "job"
    __table_args__ = make_table_args(
        comment="Attempts to send e-mails, successfully or otherwise."
    )

    # -------------------------------------------------------------------------
    # Columns
    # -------------------------------------------------------------------------

    job_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        comment="Primary key."
    )  # type: int
    recipient_id = Column(
        Integer,
        ForeignKey("recipient.recipient_id", ondelete="CASCADE"),
        nullable=False,
        comment="Which recipient? Foreign key to recipient.recipient_id"
    )
    content_id = Column(
        Integer,
        ForeignKey("content.content_id", ondelete="CASCADE"),
        nullable=False,
        comment="What content? Foreign key to content.content_id"
    )

    # -------------------------------------------------------------------------
    # Relationships
    # -------------------------------------------------------------------------

    recipient = relationship("Recipient")  # type: Recipient
    content = relationship("Content")  # type: Content

    # -------------------------------------------------------------------------
    # Jobs to be done
    # -------------------------------------------------------------------------

    @staticmethod
    def _pending_job_query(session: Session) -> CountStarSpecializedQuery:
        return (
            CountStarSpecializedQuery([Job], session=session)
            .filter(
                ~exists().select_from(SendAttempt).where(
                    and_(
                        SendAttempt.job_id == Job.job_id,
                        SendAttempt.success == True  # noqa
                    )
                )
            )
        )

    @classmethod
    def gen_jobs_to_be_done(cls, session: Session) -> Iterable["Job"]:
        query = cls._pending_job_query(session)
        for job in query.order_by(Job.job_id).all():
            yield job

    @staticmethod
    def n_completed_jobs(session: Session) -> int:
        # noinspection PyPep8
        query = (
            CountStarSpecializedQuery([Job], session=session)
            .join(SendAttempt)
            .filter(SendAttempt.success == True)  # noqa
        )
        return query.count_star()

    @classmethod
    def n_pending_jobs(cls, session: Session) -> int:
        query = cls._pending_job_query(session)
        return query.count_star()

    @classmethod
    def clear_pending_jobs(cls, session: Session) -> None:
        query = cls._pending_job_query(session)
        query.delete(synchronize_session=False)


# =============================================================================
# Attempt
# =============================================================================

class SendAttempt(Base):
    """
    Records the attempt to send an e-mail. Success means that it was accepted
    by our local server -- that doesn't imply final delivery (e.g. if the
    address is wrong).
    """
    __tablename__ = "send_attempt"
    __table_args__ = make_table_args(
        comment="Attempts to send e-mails, successfully or otherwise."
    )

    # -------------------------------------------------------------------------
    # Columns
    # -------------------------------------------------------------------------

    send_attempt_id = Column(
        Integer,
        primary_key=True,
        autoincrement=True,
        comment="Primary key."
    )  # type: int
    job_id = Column(
        Integer,
        ForeignKey("job.job_id", ondelete="CASCADE"),
        nullable=False,
        comment="Which job is this for? Foreign key to job.job_id"
    )
    config_id = Column(
        Integer,
        ForeignKey("config.config_id", ondelete="CASCADE"),
        nullable=False,
        comment="Which config was used? Foreign key to config.config_id"
    )
    when_attempt = Column(
        DateTime,
        nullable=False,
        comment="When was the attempt made?"
    )
    success = Column(
        Boolean,
        nullable=False,
        comment="Was the sending attempt successful?"
    )
    details = Column(
        Text,
        comment="If unsuccessful, error details may be here. If successful, "
                "success messages go here."
    )

    # -------------------------------------------------------------------------
    # Relationships
    # -------------------------------------------------------------------------

    job = relationship("Job")  # type: Job
    config = relationship("Config")  # type: Config
